"""Device registration module."""

# Import all device classes
from .apple.healthkit import HealthKit
from .biostrap.evo import EVO
from .coros.coros_pace_2 import CorosPace2
from .cronometer.cronometer import Cronometer
from .dexcom.pro_cgm import DexcomProCGM
from .dreem.headband_2 import DreemHeadband2
from .fitbit.fitbit_charge_4 import FitbitCharge4
from .fitbit.fitbit_charge_6 import FitbitCharge6
from .fitbit.fitbit_sense import FitbitSense
from .fitbit.google_pixel_watch import GooglePixelWatch
from .garmin.fenix_7s import Fenix7S
from .google.googlefit import GoogleFit
from .myfitnesspal.myfitnesspal import MyFitnessPal
from .nutrisense.cgm import NutrisenseCGM
from .oura.oura_ring3 import OuraRing3
from .polar.h10 import H10
from .polar.vantage import PolarVantage
from .polar.verity_sense import VeritySense
from .qualtrics.qualtrics import Qualtrics
from .registry import register_device
from .strava.strava import Strava
from .whoop.whoop_4 import Whoop4
from .withings.bodyplus import BodyPlus
from .withings.scanwatch import ScanWatch
from .withings.sleepmat import SleepMat

ALL_DEVICES = [
    HealthKit,
    EVO,
    CorosPace2,
    Cronometer,
    DexcomProCGM,
    DreemHeadband2,
    FitbitCharge4,
    FitbitCharge6,
    FitbitSense,
    GooglePixelWatch,
    Fenix7S,
    GoogleFit,
    MyFitnessPal,
    NutrisenseCGM,
    OuraRing3,
    H10,
    PolarVantage,
    VeritySense,
    Qualtrics,
    Whoop4,
    ScanWatch,
    BodyPlus,
    SleepMat,
    Strava,
]

# Register all devices
for device in ALL_DEVICES:
    register_device(device)



import numpy as np
import pandas as pd

# Storing the general data source ids for the different data types
datasourceids = {
    "steps": "derived:com.google.step_count.delta:com.google.android.gms:estimated_steps",
    "heart_rate": "derived:com.google.heart_rate.bpm:com.google.android.gms:merge_heart_rate_bpm",
    "height": "derived:com.google.height:com.google.android.gms:merge_height",
    "weight": "derived:com.google.weight:com.google.android.gms:merge_weight",
    "speed": "derived:com.google.speed:com.google.android.gms:merge_speed",
    "heart_minutes": "derived:com.google.heart_minutes:com.google.android.gms:merge_heart_minutes",
    "calories_expended": "derived:com.google.calories.expended:com.google.android.gms:merge_calories_expended",
    "sleep": "derived:com.google.sleep.segment:com.google.android.gms:merged",
    "blood_pressure": "derived:com.google.blood_pressure:com.google.android.gms:merged",
    "blood_glucose": "derived:com.google.blood_glucose:com.google.android.gms:merged",
    "activity_minutes": "derived:com.google.active_minutes:com.google.android.gms:merge_active_minutes",
    "distance": "derived:com.google.distance.delta:com.google.android.gms:merge_distance_delta",
    "oxygen_saturation": "derived:com.google.oxygen_saturation:com.google.android.gms:merged",
    "body_temperature": "derived:com.google.body.temperature:com.google.android.gms:merged",
    "menstruation": "derived:com.google.menstruation:com.google.android.gms:merged",
}


def syn_weight(x):
    return np.round(np.random.normal(70, 10), 1)


def syn_height(x):
    return np.round(np.random.normal(1.7, 0.15), 14)


weight_gen = syn_weight(1)
height_gen = syn_height(1)

# Function takes in a start date and end date and returns generated synthetic data for each data type


def create_syn_data(start_date, end_date, bucketByTime):

    # Create a list of dates between start_date and end_date
    dates = pd.date_range(start_date, end_date)

    # Create empty lists for each data type
    steps = []
    hrs = []
    weight = []
    height = []
    speed = []
    heart_minutes = []
    calories_expended = []
    sleep = []
    blood_pressure = []
    blood_glucose = []
    activity_mins = []
    distance = []
    oxygen_saturation = []
    body_temperature = []
    menstruation = []

    # Create a random device id for the synthetic data
    device_id = "".join(
        np.random.choice([*"abcdefghijklmnopqrstuvwxyz0123456789"], 8, replace=True)
    )

    iter_count = 86400000 / int(bucketByTime)

    # Iterate through each date in the list of dates
    for d in dates:
        for iter in range(int(iter_count)):
            # Convert the date to milliseconds
            startmillis = int(d.timestamp()) * 1000 + iter * int(bucketByTime)
            endmillis = int(startmillis + 86400000 / iter_count)

            # Create a resulting dictionary that we will alter for each data type
            res = {}
            res["startTimeMillis"] = str(startmillis)
            res["endTimeMillis"] = str(endmillis)

            syn_steps = max(0, int(np.random.normal(10000, 9000)))

            def max_syn_hrs(x):
                return np.round(np.random.normal(120, 20), 14)

            def syn_sleep_duration(x):
                return np.round(np.random.normal(8, 3), 1)

            def syn_speed_avg(x):
                return np.round(np.random.normal(2, 1), 1)

            def syn_speed_max(x):
                return np.round(np.random.normal(3, 1), 1)

            def syn_speed_min(x):
                return np.round(np.random.normal(1, 0.5), 1)

            def syn_heart_minutes(x):
                return np.round(np.random.normal(90, 30), 1)

            def syn_calories_expended(x):
                return np.round(np.random.normal(2000, 1000), 1)

            def syn_blood_pressure_sys(x):
                return np.round(np.random.normal(120, 20), 1)

            def syn_blood_pressure_dia(x):
                return np.round(np.random.normal(80, 20), 1)

            def syn_blood_glucose(x):
                return np.round(np.random.normal(100, 20), 1)

            def syn_activity(x):
                return np.round(np.random.normal(100, 50))

            def syn_distance(x):
                return np.round(np.random.normal(3000, 2000), 1)

            def syn_oxygen_saturation(x):
                return np.round(np.random.normal(96.5, 5), 1)

            def syn_oxygen_flow(x):
                return np.round(np.random.normal(3, 2), 1)

            def syn_body_temperature(x):
                return np.round(np.random.normal(36.5, 2), 1)

            def syn_menstruation(x):
                return np.random.choice([1, 2, 3, 4])

            # Building the resulting dictionary for steps
            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["steps"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.step_count.delta",
                                    "originDataSourceId": f"derived:com.google.step_count.delta:com.google.ios.fit:appleinc.:watch:{device_id}:top_level",
                                    "value": [{"intVal": syn_steps, "mapVal": []}],
                                }
                            ],
                        }
                    ],
                }
            ]
            steps.append(res)
            # Building the resulting dictionary for heart rate
            syn_hrs = max_syn_hrs(0)
            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["heart_rate"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.heart_rate.bpm",
                                    "originDataSourceId": f"derived:com.google.heart_rate.bpm:com.google.ios.fit:appleinc.:watch:{device_id}:top_level",
                                    "value": [
                                        {"fpVal": np.round(syn_hrs, 1), "mapVal": []},
                                        {
                                            "fpVal": np.round(
                                                syn_hrs + np.random.uniform(10, 40), 1
                                            ),
                                            "mapVal": [],
                                        },
                                        {
                                            "fpVal": np.round(
                                                syn_hrs - np.random.uniform(10, 40), 1
                                            ),
                                            "mapVal": [],
                                        },
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]

            hrs.append(res)

            # Building the resulting dictionary for sleep duration

            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["sleep"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.sleep.segment",
                                    "originDataSourceId": f"derived:com.google.sleep.segment:com.google.ios.fit:appleinc.:watch:{device_id}:top_level",
                                    "value": [
                                        {"fpVal": syn_sleep_duration(0), "mapVal": []}
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]
            sleep.append(res)

            weightGen = np.round(weight_gen + np.random.normal(-1, 1), 1)
            # Building the resulting dictionary for weight
            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["weight"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.weight",
                                    "originDataSourceId": f"derived:com.google.weight:com.google.ios.fit:appleinc.:watch:{device_id}:top_level",
                                    "value": [
                                        {"fpVal": weightGen, "mapVal": []},
                                        {"fpVal": weightGen, "mapVal": []},
                                        {"fpVal": weightGen, "mapVal": []},
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]

            weight.append(res)

            heightGen = np.round(height_gen + np.random.normal(-0.05, 0.05), 1)
            # Building the resulting dictionary for height
            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["height"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.height.summary",
                                    "originDataSourceId": "raw:com.google.height:com.google.android.apps.fitness:user_input",
                                    "value": [
                                        {"fpVal": heightGen, "mapVal": []},
                                        {"fpVal": heightGen, "mapVal": []},
                                        {"fpVal": heightGen, "mapVal": []},
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]

            height.append(res)

            # Building the resulting dictionary for speed
            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["speed"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.speed.summary",
                                    "originDataSourceId": f"derived:com.google.speed.summary:com.google.ios.fit:appleinc.:watch:{device_id}:top_level",
                                    "value": [
                                        {"fpVal": syn_speed_avg(0), "mapVal": []},
                                        {"fpVal": syn_speed_max(0), "mapVal": []},
                                        {"fpVal": syn_speed_min(0), "mapVal": []},
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]
            speed.append(res)

            # Building the resulting dictionary for blood glucose

            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["blood_glucose"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.blood_glucose.summary",
                                    "originDataSourceId": "raw:com.google.blood_glucose:com.google.android.apps.fitness:user_input",
                                    "value": [
                                        {"fpVal": syn_blood_glucose(0), "mapVal": []},
                                        {"fpVal": syn_blood_glucose(0), "mapVal": []},
                                        {"fpVal": syn_blood_glucose(0), "mapVal": []},
                                        {"intVal": 2, "mapVal": []},
                                        {"intVal": 3, "mapVal": []},
                                        {"intVal": 2, "mapVal": []},
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]

            blood_glucose.append(res)

            # Building the resulting dictionary for blood pressure

            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["blood_pressure"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.blood_pressure.summary",
                                    "originDataSourceId": "raw:com.google.blood_pressure:com.google.android.apps.fitness:user_input",
                                    "value": [
                                        {
                                            "fpVal": syn_blood_pressure_sys(0),
                                            "mapVal": [],
                                        },
                                        {
                                            "fpVal": syn_blood_pressure_dia(0),
                                            "mapVal": [],
                                        },
                                        {
                                            "fpVal": syn_blood_pressure_dia(0),
                                            "mapVal": [],
                                        },
                                        {
                                            "fpVal": syn_blood_pressure_dia(0),
                                            "mapVal": [],
                                        },
                                        {
                                            "fpVal": syn_blood_pressure_dia(0),
                                            "mapVal": [],
                                        },
                                        {"intVal": 2, "mapVal": []},
                                        {"intVal": 2, "mapVal": []},
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]
            blood_pressure.append(res)

            # Building the resulting dictionary for distance
            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["distance"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.distance.delta",
                                    "originDataSourceId": f"derived:com.google.distance.delta:com.google.ios.fit:appleinc.:watch:{device_id}:top_level",
                                    "value": [
                                        {"fpVal": abs(syn_distance(0)), "mapVal": []},
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]
            distance.append(res)

            # Building the resulting dictionary for heart minutes
            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["heart_minutes"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.heart_minutes.summary",
                                    "originDataSourceId": f"derived:com.google.heart_minutes.summary:com.google.ios.fit:appleinc.:watch:{device_id}:top_level",
                                    "value": [
                                        {"fpVal": syn_heart_minutes(0), "mapVal": []},
                                        {"fpVal": syn_heart_minutes(0), "mapVal": []},
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]
            heart_minutes.append(res)

            # Building the resulting dictionary for calories

            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["calories_expended"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.calories.expended",
                                    "originDataSourceId": f"derived:com.google.calories.expended:com.google.ios.fit:appleinc.:watch:{device_id}:top_level",
                                    "value": [
                                        {
                                            "fpVal": syn_calories_expended(0),
                                            "mapVal": [],
                                        },
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]

            calories_expended.append(res)

            # Building the resulting dictionary for activity_minutes
            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["activity_minutes"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.activity_minutes",
                                    "originDataSourceId": f"derived:com.google.activity_minutes.summary:com.google.ios.fit:appleinc.:watch:{device_id}:top_level",
                                    "value": [{"fpVal": syn_activity(0), "mapVal": []}],
                                }
                            ],
                        }
                    ],
                }
            ]

            activity_mins.append(res)

            # Building the resulting dictionary for menstrual cycle

            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["menstruation"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.menstrual_cycle",
                                    "originDataSourceId": "raw:com.google.menstrual_cycle:com.google.android.apps.fitness:user_input",
                                    "value": [
                                        {
                                            "intVal": int(syn_menstruation(0)),
                                            "mapVal": [],
                                        }
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]
            menstruation.append(res)

            # Building the resulting dictionary for body temperature

            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["body_temperature"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.body.temperature",
                                    "originDataSourceId": "raw:com.google.body.temperature:com.google.android.apps.fitness:user_input",
                                    "value": [
                                        {
                                            "fpVal": syn_body_temperature(0),
                                            "mapVal": [],
                                        },
                                        {
                                            "fpVal": syn_body_temperature(0),
                                            "mapVal": [],
                                        },
                                        {
                                            "fpVal": syn_body_temperature(0),
                                            "mapVal": [],
                                        },
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]
            body_temperature.append(res)

            # Building the resulting dictionary for oxygen saturation

            res = [
                {
                    "startTimeMillis": startmillis,
                    "endTimeMillis": endmillis,
                    "dataset": [
                        {
                            "dataSourceId": datasourceids["oxygen_saturation"],
                            "point": [
                                {
                                    "startTimeNanos": str(startmillis * 1000000),
                                    "endTimeNanos": str(endmillis * 1000000),
                                    "dataTypeName": "com.google.oxygen_saturation",
                                    "originDataSourceId": "raw:com.google.oxygen_saturation:com.google.android.apps.fitness:user_input",
                                    "value": [
                                        {
                                            "fpVal": syn_oxygen_saturation(0),
                                            "mapVal": [],
                                        },
                                        {
                                            "fpVal": syn_oxygen_saturation(0),
                                            "mapVal": [],
                                        },
                                        {
                                            "fpVal": syn_oxygen_saturation(0),
                                            "mapVal": [],
                                        },
                                        {"fpVal": syn_oxygen_flow(0), "mapVal": []},
                                        {"fpVal": syn_oxygen_flow(0), "mapVal": []},
                                        {"fpVal": syn_oxygen_flow(0), "mapVal": []},
                                        {"intVal": 1, "mapVal": []},
                                        {"intVal": 1, "mapVal": []},
                                        {"intVal": 1, "mapVal": []},
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]
            oxygen_saturation.append(res)

    # Retuning the resulting lists
    return (
        steps,
        hrs,
        weight,
        height,
        speed,
        heart_minutes,
        calories_expended,
        sleep,
        blood_pressure,
        blood_glucose,
        activity_mins,
        distance,
        oxygen_saturation,
        body_temperature,
        menstruation,
    )



import json
import os

import numpy as np
import pandas as pd


def create_syn_data(start_date, end_date):

    # Create a list of dates between start_date and end_date
    dates = pd.date_range(start_date, end_date, freq="D")

    # Create empty lists to store data
    goals = []
    daily_summary = []
    strength_exercises = []
    cardio_exercises = []
    breakfast = []
    lunch = []
    dinner = []
    snacks = []

    # Functions to generate synthetic data
    def syn_calories(x):
        return np.round(np.random.normal(2500, 200), 1)

    def syn_carbs(x):
        return np.round(np.random.normal(250, 75), 1)

    def syn_fat(x):
        return np.round(np.random.normal(75, 25), 1)

    def syn_protein(x):
        return np.round(max(0.0, np.random.normal(100, 33)), 1)

    def syn_sodium(x):
        return np.round(np.random.normal(2300, 500), 1)

    def syn_sugar(x):
        return np.round(np.random.normal(75, 25), 1)

    path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "myfitnesspal_syn_data.json")
    )

    # Loading data from our json file
    data = json.load(open(path, "r"))

    # wearipedia/devices/myfitnesspal/myfitnesspal_syn_data.json

    # Scraped a list of all exercises from MyFitnessPal and used GPT3 to seperate them into strength and cardio exercises
    exercises_cardio = data["exercises_cardio"]
    # List of exercises that are strength exercises scraped from myfitnesspal.com
    exercises_strength = data["exercises_strength"]

    # List of top 15 breakfast foods from the USDA
    breakfast_foods = data["breakfast_foods"]

    # List of top 15 lunch foods from the USDA
    lunch_foods = data["lunch_foods"]
    # List of top 15 dinner foods from the USDA
    dinner_foods = data["dinner_foods"]
    # List of top 15 snack foods from the USDA
    snack_foods = data["snack_foods"]

    def create_food_item(item_name, food_item):
        return [
            {"day": pd.Timestamp(day)},
            {
                "name": item_name,
                "nutrition_information": {
                    "calories": float(food_item["calories"]),
                    "carbohydrates": float(food_item["carbohydrates"]),
                    "fat": float(food_item["fat"]),
                    "protein": float(food_item["protein"]),
                    "sodium": float(food_item["sodium"]),
                    "sugar": float(food_item["sugar"]),
                },
                "totals": {
                    "calories": float(food_item["calories"]),
                    "carbohydrates": float(food_item["carbohydrates"]),
                    "fat": float(food_item["fat"]),
                    "protein": float(food_item["protein"]),
                    "sodium": float(food_item["sodium"]),
                    "sugar": float(food_item["sugar"]),
                },
            },
        ]

    for day in dates:

        # Creating a randomly generated summary of the day's food goal
        goals.append(
            {
                "calories": syn_calories(day),
                "carbohydrates": syn_carbs(day),
                "fat": syn_fat(day),
                "protein": syn_protein(day),
                "sodium": syn_sodium(day),
                "sugar": syn_sugar(day),
                "date": pd.Timestamp(day),
            }
        )

        # Creating a randomly generated summary of the day's food intake
        daily_summary.append(
            {
                "calories": syn_calories(day),
                "carbohydrates": syn_carbs(day),
                "fat": syn_fat(day),
                "protein": syn_protein(day),
                "sodium": syn_sodium(day),
                "sugar": syn_sugar(day),
                "date": pd.Timestamp(day),
            }
        )

        # Creating a randomly generated list of cardio exercises for the day
        cardio = [{"day": pd.Timestamp(day)}]

        # We will randomly select between 1 and 3 cardio exercises
        cardio_count = np.random.randint(1, 3)

        # We will randomly select between 1 and 3 cardio exercises
        random_exercises = np.random.choice(
            exercises_cardio, cardio_count, replace=False
        )

        # Adding each exercise to the list of cardio exercises for that day
        for exercise in random_exercises:
            minutes = np.random.randint(10, 60)
            syn_exercise = {
                "name": exercise,
                "nutrition_information": {
                    "minutes": minutes,
                    "calories burned": minutes * max(2, np.random.uniform(3, 5)),
                },
            }

            # Adding the exercise to the list of cardio exercises for that day
            cardio.append(syn_exercise)

        # Adding the list of cardio exercises for that day to the list of all cardio exercises
        cardio_exercises.append(cardio)

        # Creating a randomly generated list of strength exercises for the day
        strength = []

        # Adding the date to the list of strength exercises for that day
        strength.append({"date": pd.Timestamp(day)})

        # We will randomly select between 1 and 10 strength exercises
        exercise_count = np.random.randint(1, 10)

        # We will randomly select between 1 and 10 strength exercises
        random_exercises = np.random.choice(
            exercises_strength, exercise_count, replace=False
        )

        # Adding each exercise to the list of strength exercises for that day
        for exercise in random_exercises:
            syn_exercise = {
                "name": exercise,
                "nutrition_information": {
                    "sets": float(np.random.randint(2, 5)),
                    "reps/set": float(np.round(np.random.uniform(10, 2))),
                    "weight/set": float(np.random.randint(5, 100)),
                },
            }
            strength.append(syn_exercise)
        strength_exercises.append(strength)

        # Creating a randomly generated list of breakfast foods for the day
        breakfast_item_name = np.random.choice(list(breakfast_foods.keys()))
        breakfast_item = breakfast_foods[breakfast_item_name]

        # Adding the breakfast food to the list of breakfast foods for that day
        breakfast.append(create_food_item(breakfast_item_name, breakfast_item))

        # Creating a randomly generated list of lunch foods for the day
        lunch_item_name = np.random.choice(list(lunch_foods.keys()))
        lunch_item = lunch_foods[lunch_item_name]

        # Adding the lunch food to the list of lunch foods for that day
        lunch.append(create_food_item(lunch_item_name, lunch_item))

        # Creating a randomly generated list of dinner foods for the day
        dinner_item_name = np.random.choice(list(dinner_foods.keys()))
        dinner_item = dinner_foods[dinner_item_name]

        # Adding the dinner food to the list of dinner foods for that day
        dinner.append(create_food_item(dinner_item_name, dinner_item))

        # Creating a randomly generated list of snack foods for the day
        snack_item_name = np.random.choice(list(snack_foods.keys()))
        snack_item = snack_foods[snack_item_name]

        # Adding the snack food to the list of snack foods for that day
        snacks.append(create_food_item(snack_item_name, snack_item))

    return (
        goals,
        daily_summary,
        cardio_exercises,
        strength_exercises,
        breakfast,
        lunch,
        dinner,
        snacks,
    )



import requests

__all__ = ["fetch_real_data"]


def call_api_version_2(
    url: str,
    access_token,
    start_date,
    end_date,
    start_date_col,
    end_date_col,
    call: str = "GET",
):
    """
    Second version of the api, the only supported API version as of 1/24/25
    """
    headers = {"Authorization": "Bearer " + access_token}
    params = {start_date_col: start_date, end_date_col: end_date}

    response = requests.request(call, url=url, headers=headers, params=params)

    # Handle specific HTTP status codes
    if response.status_code != 200:
        error_msg = f"{response.status_code}"
        try:
            error_detail = response.json()
            if isinstance(error_detail, dict):
                error_msg = f"{error_msg} - {error_detail.get('detail', '')}"
        except:
            pass

        raise Exception("Request failed with error: " + error_msg)
    return response.json()


def fetch_real_data(data_type, access_token, start_date, end_date):
    """Main function for fetching real data from the Oura Ring API.

    :param start_date: the start date represented as a string in the format "YYYY-MM-DD"
    :param end_date: the end date represented as a string in the format "YYYY-MM-DD"
    :param data_type: the type of data to fetch, one of "Personal Info", "Heart Rate", "Sessions", "Tags", "Workouts", "Daily Sleep", "Daily Activity", "Daily Readiness", "Ideal Bedtime"
    :param access_token: access token for the API
    :return: the data fetched from the API according to the inputs
    :rtype: List
    """

    start_date_col = "start_date"
    end_date_col = "end_date"
    if data_type == "heart_rate":
        endpoint = "https://api.ouraring.com/v2/usercollection/heartrate"
        start_date_col = "start_datetime"
        end_date_col = "end_datetime"
        start_date = start_date + "T00:00:00-23:59"
        end_date = end_date + "T00:00:00-23:59"
    elif data_type == "personal_info":
        endpoint = "https://api.ouraring.com/v2/usercollection/personal_info"
    elif data_type == "session":
        endpoint = "https://api.ouraring.com/v2/usercollection/session"
    elif data_type == "enhanced_tag":
        # Tag is deprecated, and the Oura website recommends transitioning to enhanced_tag
        # (https://cloud.ouraring.com/v2/docs#tag/Tag-Routes)
        endpoint = "https://api.ouraring.com/v2/usercollection/enhanced_tag"
    elif data_type == "workout":
        endpoint = "https://api.ouraring.com/v2/usercollection/workout"
    elif data_type == "daily_activity":
        endpoint = "https://api.ouraring.com/v2/usercollection/daily_activity"
    elif data_type == "daily_sleep":
        endpoint = "https://api.ouraring.com/v2/usercollection/daily_sleep"
    elif data_type == "sleep":
        endpoint = "https://api.ouraring.com/v2/usercollection/sleep"
    elif data_type == "readiness":
        endpoint = "https://api.ouraring.com/v2/usercollection/daily_readiness"
    elif data_type == "ideal_sleep_time":
        endpoint = "https://api.ouraring.com/v2/usercollection/sleep_time"

    response = call_api_version_2(
        endpoint, access_token, start_date, end_date, start_date_col, end_date_col
    )

    if data_type == "personal_info":
        return [response]
    return response["data"]



import pandas as pd

# Functions to generate data
# These functions are used to generate data that closely mimic the
# data that is generated by the MyFitnessPal API


def goal_generator(self, days):
    # creating an empty list to store the goals
    goals = []

    # for each day in the date range, we get the goals for that day
    for day in days:
        # using the client, we get the goals for the day
        res = self.client.get_date(int(day.year), int(day.month), int(day.day)).goals
        # we add the date to the goals
        res["date"] = day
        # we append the goals to the list
        goals.append(res)
    # we return the list of goals
    return goals


def daily_summary_generator(self, days):
    # creating an empty list to store the daily summary
    summary = []

    # for each day in the date range, we get the daily summary for that day
    for day in days:
        # using the client, we get the daily summary for the day
        res = self.client.get_date(int(day.year), int(day.month), int(day.day)).totals
        # we add the date to the daily summary
        res["date"] = day
        # we append the daily summary to the list
        summary.append(res)

    # we return the list of daily summary
    return summary


def cardio_generator(self, days):
    # creating an empty list to store the cardio exercises
    cardio = []

    # for each day in the date range, we get the cardio exercises for that day
    for day in days:
        # using the client, we get the cardio exercises for the day
        res = (
            self.client.get_date(int(day.year), int(day.month), int(day.day))
            .exercises[0]
            .get_as_list()
        )
        # we add the date to the cardio exercises
        res = [{"day": day}] + res
        # we append the cardio exercises to the list
        cardio.append(res)
    # we return the list of cardio exercises
    return cardio


def strength_generator(self, days):
    # creating an empty list to store the strength exercises
    strength = []

    # for each day in the date range, we get the strength exercises for that day
    for day in days:
        # using the client, we get the strength exercises for the day
        res = (
            self.client.get_date(int(day.year), int(day.month), int(day.day))
            .exercises[1]
            .get_as_list()
        )
        # we add the date to the strength exercises
        res = [{"day": day}] + res
        # we append the strength exercises to the list
        strength.append(res)
    # we return the list of strength exercises
    return strength


def food_fetcher(client, type, days):
    # creating an empty list to store the food items
    food = []

    # for each day in the date range, we get the food items for that day
    for day in days:
        # using the client, we get the food items for the day
        res = (
            client.get_date(int(day.year), int(day.month), int(day.day))
            .meals[type]
            .get_as_list()
        )

        # if the length of the list is 0, skip the totals
        if len(res) == 0:
            res.append({"date": day})

        # if the length of the list is not 0, we add the date and the totals to the list
        else:
            res[0]["date"] = day
            res[0]["totals"] = (
                client.get_date(int(day.year), int(day.month), int(day.day))
                .meals[type]
                .totals
            )

        # we append the food items to the list
        food.append(res)
    # we return the list of food items
    return food


def breakfast_generator(self, days):
    # creating an empty list to store the breakfast foods
    return food_fetcher(self.client, 0, days)


def lunch_generator(self, days):
    # creating an empty list to store the lunch foods
    return food_fetcher(self.client, 1, days)


def dinner_generator(self, days):
    # creating an empty list to store the dinner foods
    return food_fetcher(self.client, 2, days)


def snacks_generator(self, days):
    # creating an empty list to store the snacks foods
    return food_fetcher(self.client, 3, days)


# This is the function that will be used to fetch data from MyFitnessPal
def fetch_real_data(self, start_date, end_date, data_type):

    # if the client is not set, we need to login
    if self.client == None:
        raise Exception("Not Authenticated, login and try again")

    # creating the date range from the start and end dates
    days = pd.date_range(start_date, end_date, freq="D")

    # if the data type is goals, we need to get the goals for each day using the client
    if data_type == "goals":
        return goal_generator(self, days)

    # if the data type is daily_summary, we need to get the daily summary for each day using the client
    elif data_type == "daily_summary":
        return daily_summary_generator(self, days)

    # if the data type is exercises_cardio, we need to get the cardio exercises for each day using the client
    if data_type == "exercises_cardio":
        return cardio_generator(self, days)

    # if the data type is exercises_strength, we need to get the strength exercises for each day using the client
    if data_type == "exercises_strength":
        return strength_generator(self, days)

    # if the data type is breakfast, we need to get the other exercises for each day using the client
    if data_type == "breakfast":
        return breakfast_generator(self, days)

    # if the data type is lunch, we need to get the lunch foods for each day using the client
    if data_type == "lunch":
        return lunch_generator(self, days)

    # if the data type is dinner, we need to get the dinner foods for each day using the client
    if data_type == "dinner":
        return dinner_generator(self, days)

    # if the data type is snacks, we need to get the snacks for each day using the client
    if data_type == "snacks":
        return snacks_generator(self, days)



import collections
from datetime import datetime, time, timedelta

from ...utils import bin_search, seed_everything
from ..device import BaseDevice
from .coros_pace_2_fetch import fetch_real_data
from .coros_pace_2_gen import create_syn_data


class CorosPace2(BaseDevice):
    """This device allows you to work with data from the `Fitbit charge  <(https://www.fitbit.com/global/au/products/trackers/charge4)>`_ device.
    Available datatypes for this device are:

    * `steps`: steps data
    * `exercise_time`: exercise time data
    * `heart_rate`: heart rate data
    * `sports`: sports data
    * `sleep`: sleep data
    * `active_energy`: active energy data

    :param seed: random seed for synthetic data generation, defaults to 0
    :type seed: int, optional
    :param synthetic_start_date: start date for synthetic data generation, defaults to "2022-03-01"
    :type synthetic_start_date: str, optional
    :param synthetic_end_date: end date for synthetic data generation, defaults to "2022-06-17"
    :type synthetic_end_date: str, optional
    """

    name = "coros/coros_pace_2"

    def __init__(
        self,
        seed=0,
        synthetic_start_date="2022-03-01",
        synthetic_end_date="2022-06-17",
    ):

        params = {
            "seed": seed,
            "synthetic_start_date": synthetic_start_date,
            "synthetic_end_date": synthetic_end_date,
        }

        self._initialize_device_params(
            [
                "steps",
                "exercise_time",
                "heart_rate",
                "sports",
                "sleep",
                "active_energy",
            ],
            params,
            {
                "seed": 0,
                "synthetic_start_date": "2022-03-01",
                "synthetic_end_date": "2022-06-17",
            },
        )

    def _default_params(self):
        params = {
            "start_date": "2022-04-24",
            "end_date": "2022-04-28",
        }

        return params

    def _filter_synthetic(self, data, data_type, params):

        date_format = "%Y-%m-%d"
        date1 = datetime.strptime(self.init_params["synthetic_start_date"], date_format)
        date2 = datetime.strptime(params["start_date"], date_format)

        date3 = datetime.strptime(self.init_params["synthetic_end_date"], date_format)
        date4 = datetime.strptime(params["end_date"], date_format)

        delta1 = date2 - date1
        delta2 = date3 - date4

        num_days_start = delta1.days
        num_days_end = delta2.days

        return data[num_days_start : -num_days_end + 1]

    def _get_real(self, data_type, params):

        data = fetch_real_data(
            data_type,
            self.user,
            start_date=self.init_params["start_date"],
            end_date=self.init_params["end_date"],
        )
        return data

    def _gen_synthetic(self):

        syn_data = create_syn_data(
            self.init_params["synthetic_start_date"],
            self.init_params["synthetic_end_date"],
        )
        self.steps = syn_data["steps"]
        self.exercise_time = syn_data["exercise_time"]
        self.heart_rate = syn_data["heart_rate"]
        self.sports = syn_data["sports"]
        self.sleep = syn_data["sleep"]
        self.active_energy = syn_data["active_energy"]

    def _authenticate(self):
        # authenticate this device against API
        print(
            "Get the access token here: https://colab.research.google.com/drive/1cnVktqcPXPtGS6JcdHVVCqIC3nngfxhm?usp=sharing"
        )
        token = input("Enter your access token: ")
        self.user = token



from datetime import datetime, time, timedelta

from ...utils import bin_search, seed_everything
from ..device import BaseDevice
from .oura_ring3_authenticate import oura_token
from .oura_ring3_fetch import fetch_real_data
from .oura_ring3_gen import create_syn_data


class OuraRing3(BaseDevice):
    """This device allows you to work with data from the `Oura Ring 3  <(https://ouraring.com/)>`_ device.
    Available datatypes for this device are:

    * `heart_rate`: heart_rate
    * `personal_info`: personal information
    * `session`: data on user engagement with sessions in the Oura app
    * `enhanced_tag`: user tags on data
    * `workout`: workout sessions
    * `daily_activity`: daily activity data
    * `daily_sleep`: daily sleep data including scores per day
    * `sleep`: all logged sleep activities
    * `readiness`: readiness data
    * `ideal_sleep_time`: ideal bedtime data

    :param seed: random seed for synthetic data generation, defaults to 0
    :type seed: int, optional
    :param synthetic_start_date: start date for synthetic data generation, defaults to "2022-03-01"
    :type synthetic_start_date: str, optional
    :param synthetic_end_date: end date for synthetic data generation, defaults to "2022-06-17"
    :type synthetic_end_date: str, optional
    """

    name = "oura/oura_ring3"

    def __init__(
        self,
        seed=0,
        synthetic_start_date="2022-03-01",
        synthetic_end_date="2022-06-17",
    ):

        params = {
            "seed": seed,
            "synthetic_start_date": synthetic_start_date,
            "synthetic_end_date": synthetic_end_date,
        }

        self._initialize_device_params(
            [
                "heart_rate",
                "personal_info",
                "session",
                "enhanced_tag",
                "workout",
                "daily_activity",
                "daily_sleep",
                "sleep",
                "readiness",
                "ideal_sleep_time",
            ],
            params,
            {
                "seed": 0,
                "synthetic_start_date": "2022-03-01",
                "synthetic_end_date": "2022-06-17",
            },
        )

    def _default_params(self):
        params = {
            "seed": 0,
            "start_date": "2022-04-24",
            "end_date": "2022-04-28",
        }

        return params

    def _filter_synthetic(self, data, data_type, params):

        date_format = "%Y-%m-%d"
        date1 = datetime.strptime(self.init_params["synthetic_start_date"], date_format)
        date2 = datetime.strptime(params["start_date"], date_format)

        date3 = datetime.strptime(self.init_params["synthetic_end_date"], date_format)
        date4 = datetime.strptime(params["end_date"], date_format)

        delta1 = date2 - date1
        delta2 = date3 - date4

        num_days_start = delta1.days
        num_days_end = delta2.days

        return data[num_days_start : -num_days_end + 1]

    def _get_real(self, data_type, params):

        data = fetch_real_data(
            data_type,
            self.user,
            start_date=params["start_date"],
            end_date=params["end_date"],
        )
        return data

    def _gen_synthetic(self):

        syn_data = create_syn_data(
            self.init_params["seed"],
            self.init_params["synthetic_start_date"],
            self.init_params["synthetic_end_date"],
        )
        self.personal_info = []
        self.heart_rate = syn_data["heart_rate"]
        self.session = syn_data["session"]
        self.enhanced_tag = syn_data["enhanced_tag"]
        self.workout = syn_data["workout"]
        self.daily_activity = syn_data["daily_activity"]
        self.daily_sleep = syn_data["daily_sleep"]
        self.sleep = syn_data["sleep"]
        self.readiness = syn_data["readiness"]
        self.ideal_sleep_time = syn_data["ideal_sleep_time"]

    def _authenticate(self, token=""):
        if token == "":
            self.user = oura_token()
        else:
            self.user = token



from datetime import datetime, timedelta

import requests

__all__ = ["fetch_real_data"]


def call_API(url: str, access_token: str, call: str = "GET"):
    headers = {"Authorization": "Bearer " + access_token}
    response = requests.request(call, url=url, headers=headers)
    # Handle specific HTTP status codes
    if response.status_code != 200:
        error_msg = f"{response.status_code}"
        try:
            error_detail = response.json()
            if isinstance(error_detail, dict):
                error_msg = f"{error_msg} - {error_detail}"
        except:
            pass

        raise Exception("Request failed with error: " + error_msg)
    return response.json()


def fetch_real_data(data_type, access_token, start_date=None, end_date=None):
    """Main function for fetching real data from the Fitbit API.

    :param start_date: the start date represented as a string in the format "YYYY-MM-DD"
    :type start_date: str
    :param end_date: the end date represented as a string in the format "YYYY-MM-DD"
    :type end_date: str
    :param data_type: the type of data to fetch, one of "sleep", "steps","minutesVeryActive", "minutesLightlyActive", "minutesFairlyActive", "distance", "minutesSedentary", "heart_rate_day", "hrv", "distance_day", "breath_rate"
    :type data_type: str
    :param access_token: access token for the API
    :type api: str
    :return: the data fetched from the API according to the inputs
    :rtype: List
    """
    categories = {
        "sleep": {
            "url": f"https://api.fitbit.com/1.2/user/-/sleep/date/{start_date}/{end_date}.json"
        },
        "steps": {
            "url": f"https://api.fitbit.com/1/user/-/activities/steps/date/{start_date}/{end_date}.json"
        },
        "minutesVeryActive": {
            "url": f"https://api.fitbit.com/1/user/-/activities/minutesVeryActive/date/{start_date}/{end_date}.json"
        },
        "minutesFairlyActive": {
            "url": f"https://api.fitbit.com/1/user/-/activities/minutesFairlyActive/date/{start_date}/{end_date}.json"
        },
        "minutesLightlyActive": {
            "url": f"https://api.fitbit.com/1/user/-/activities/minutesLightlyActive/date/{start_date}/{end_date}.json"
        },
        "distance": {
            "url": f"https://api.fitbit.com/1/user/-/activities/distance/date/{start_date}/{end_date}.json"
        },
        "minutesSedentary": {
            "url": f"https://api.fitbit.com/1/user/-/activities/minutesSedentary/date/{start_date}/{end_date}.json"
        },
        "heart_rate_day": {
            "url": f"https://api.fitbit.com/1/user/-/activities/heart/date/{start_date}/{end_date}/1d.json"
        },
        "hrv": {
            "url": f"https://api.fitbit.com/1/user/-/hrv/date/{start_date}/{end_date}.json"
        },
        "distance_day": {
            "url": f"https://api.fitbit.com/1/user/-/activities/distance/date/{start_date}/{end_date}/1d.json"
        },
    }

    arr = []

    if "intraday" in data_type:
        current_date = datetime.strptime(start_date, "%Y-%m-%d")
        end = datetime.strptime(end_date, "%Y-%m-%d")

        # Iterate through dates until end date is reached
        while current_date <= end:
            current_date
            current_date += timedelta(days=1)
            current_date_str = current_date.strftime("%Y-%m-%d")
            intraday_categories = {
                "intraday_breath_rate": {
                    "url": f"https://api.fitbit.com/1/user/-/br/date/{current_date_str}/all.json"
                },
                "intraday_active_zone_minute": {
                    "url": f"https://api.fitbit.com/1/user/-/activities/active-zone-minutes/date/{current_date_str}/1d/1min.json"
                },
                "intraday_activity": {
                    "url": f"https://api.fitbit.com/1/user/-/activities/steps/date/{current_date_str}/1d/1min.json"
                },
                "intraday_heart_rate": {
                    "url": f"https://api.fitbit.com/1/user/-/activities/heart/date/{current_date_str}/1d/1sec.json"
                },
                "intraday_hrv": {
                    "url": f"https://api.fitbit.com/1/user/-/hrv/date/{current_date_str}/all.json"
                },
                "intraday_spo2": {
                    "url": f"https://api.fitbit.com/1/user/-/spo2/date/{current_date_str}/all.json"
                },
            }
            response = call_API(
                url=intraday_categories[data_type]["url"], access_token=access_token
            )
            arr.append(response)
    else:
        response = call_API(url=categories[data_type]["url"], access_token=access_token)
    arr.append(response)

    return arr



from datetime import datetime

import numpy as np
import pandas as pd

from ...devices.device import BaseDevice
from ...utils import bin_search, seed_everything
from .apple_gen import create_syn_data


class HealthKit(BaseDevice):
    """This device allows you to work with data from `Apple HealthKit <https://www.apple.com/ios/health/>`_ .
    Available datatypes for this device are:

    * `dates`: a list of consecutive dates

    * `steps`: a sibling list to `dates` that contains step data for each day

    * `hrs`: a sibling list to `dates` that contains heart rate data for each day

    :param seed: random seed for synthetic data generation, defaults to 0
    :type seed: int, optional
    :param synthetic_start_date: start date for synthetic data generation, defaults to "2022-03-01"
    :type synthetic_start_date: str, optional
    :param synthetic_end_date: end date for synthetic data generation, defaults to "2022-06-17"
    :type synthetic_end_date: str, optional
    :param use_cache: decide whether to cache the credentials, defaults to True
    :type use_cache: bool, optional
    """

    name = "apple/healthkit"

    def __init__(
        self,
        seed=0,
        synthetic_start_date="2022-03-01",
        synthetic_end_date="2022-06-17",
    ):

        params = {
            "seed": seed,
            "synthetic_start_date": synthetic_start_date,
            "synthetic_end_date": synthetic_end_date,
        }

        self._initialize_device_params(
            ["dates", "steps", "hrs"],
            params,
            {
                "seed": 0,
                "synthetic_start_date": "2022-03-01",
                "synthetic_end_date": "2022-06-17",
            },
        )

    def _default_params(self):
        return {
            "start_date": self.init_params["synthetic_start_date"],
            "end_date": self.init_params["synthetic_end_date"],
        }

    def _filter_synthetic(self, data, data_type, params):
        # Here we just return the data we've already generated,
        # but index into it based on the params. Specifically, we
        # want to return the data between the start and end dates.

        date_str_to_obj = lambda x: datetime.strptime(x, "%Y-%m-%d")

        # get the indices by subtracting against the start of the synthetic data
        synthetic_start = date_str_to_obj(self.init_params["synthetic_start_date"])

        start_idx = (date_str_to_obj(params["start_date"]) - synthetic_start).days
        end_idx = (date_str_to_obj(params["end_date"]) - synthetic_start).days

        return data[start_idx:end_idx]

    def _gen_synthetic(self):
        seed_everything(self.init_params["seed"])

        self.dates, self.steps, self.hrs = create_syn_data(
            self.init_params["synthetic_start_date"],
            self.init_params["synthetic_end_date"],
        )



import copy
import datetime
from collections import defaultdict
from threading import Thread

import numpy as np
import pandas as pd
from tqdm import tqdm


def gen_data(seed, start_date, end_date):
    """Main function for creating synthetic heart rate data for the H10.

    :param start_date: the start date represented as a string in the format "YYYY-MM-DD"
    :type start_date: str
    :param end_date: the end date represented as a string in the format "YYYY-MM-DD"
    :type end_date: str
    :return: a tuple of dictionary with keys the training session dates and values a dictionary with keys RR< heart_rates, calories, and minutes
    :rtype: tuple(Dict[str: list, str: list], Dict[str: Dict[str: list, str: int, str: int]])
    """
    sessions = np.datetime64(end_date) - np.datetime64(start_date)
    durations = (45, 60)  # minutes

    def gen_all(rr_result, hr_result, index):
        local_rng = np.random.RandomState(seed + index)

        # simulate skip day
        if local_rng.uniform(low=0, high=1, size=(1,))[0] > 0.8:
            return

        # day that you workout
        day = np.datetime64(start_date) + np.timedelta64(index, "D")
        duration = int(
            local_rng.uniform(low=durations[0], high=durations[1], size=(1,))[0]
        )

        # RR data
        rr_list = []
        c_rr = local_rng.uniform(low=400, high=2000, size=(1,))[0]

        # heart rate data
        hrate = []
        start_rate = local_rng.uniform(low=70, high=110, size=(1,))[0]

        for _ in range(duration * 60):

            # heart rate data
            hrate.append(start_rate)
            added = local_rng.normal(scale=1) + 0.01 * (160 / start_rate)
            if start_rate < 50:
                added = abs(added)
            elif start_rate > 190:
                added = -1 * abs(added)
            start_rate += added

            # RR data
            rr_list.append(c_rr)
            added = local_rng.normal(scale=1) + 0.01 * (1000 / c_rr)
            if c_rr < 400:
                added = abs(added)
            elif c_rr > 2000:
                added = -1 * abs(added)
            c_rr += added

        # save heart rate data
        hr_result[pd.to_datetime(day).strftime("%Y-%m-%d")]["heart_rates"] = hrate
        hr_result[pd.to_datetime(day).strftime("%Y-%m-%d")]["calories"] = int(
            local_rng.uniform(low=200, high=500, size=(1,))[0]
        )
        hr_result[pd.to_datetime(day).strftime("%Y-%m-%d")]["minutes"] = duration

        # save RR data
        # create a list of timestamps
        cur_time = datetime.datetime.strptime("00:00:00.0", "%H:%M:%S.%f")
        date_list = []

        for interval in rr_list:
            date_list.append(cur_time)
            cur_time = cur_time + datetime.timedelta(milliseconds=interval)

        rr_result[pd.to_datetime(day).strftime("%Y-%m-%d")]["rr"] = rr_list
        rr_result[pd.to_datetime(day).strftime("%Y-%m-%d")]["time"] = date_list

    threads = []
    hr_result = defaultdict(dict)
    rr_result = defaultdict(dict)
    for i in tqdm(range(int(sessions / np.timedelta64(1, "D")))):
        new_thread = Thread(target=gen_all, args=(rr_result, hr_result, i))
        threads.append(new_thread)

    # start threads
    for thread in threads:
        thread.start()

    # wait for all threads to terminate
    for thread in tqdm(threads):
        thread.join()

    return rr_result, hr_result



# utility scripts to extract via Withings API

####################################
# Withings ScanWatch notebook code #
####################################

# @title Enter start and end dates
start_date = "2020-05-20"  # @param {type:"date"}
end_date = "2022-07-20"  # @param {type:"date"}

import json
import time
import urllib
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import requests
from tqdm import tqdm

# import july
# from july.utils import date_range


num_to_description = {
    1: "Weight (kg)",
    4: "Height (meter)",
    5: "Fat Free Mass (kg)",
    6: "Fat Ratio (%)",
    8: "Fat Mass Weight (kg)",
    9: "Diastolic Blood Pressure (mmHg)",
    10: "Systolic Blood Pressure (mmHg)",
    11: "Heart Pulse (bpm) - only for BPM and scale devices",
    12: "Temperature (celsius)",
    54: "SP02 (%)",
    71: "Body Temperature (celsius)",
    73: "Skin Temperature (celsius)",
    76: "Muscle Mass (kg)",
    77: "Hydration (kg)",
    88: "Bone Mass (kg)",
    91: "Pulse Wave Velocity (m/s)",
    123: "VO2 max is a numerical measurement of your body’s ability to consume oxygen (ml/min/kg).",
    135: "QRS interval duration based on ECG signal",
    136: "PR interval duration based on ECG signal",
    137: "QT interval duration based on ECG signal",
    138: "Corrected QT interval duration based on ECG signal",
    139: "Atrial fibrillation result from PPG",
}

NUM_RETRIES = 3


def fetch_all_wrapper(endpoint_url, data, headers, arr_key, parse_data=lambda x: x):
    # wrapper around public API that retrieves arbitrarily large # of
    # records, since there is a restriction of # of records per API response
    # NOTES:
    # out['body'][arr_key] is concatenated across several requests
    # parse_data is a function that parses the returned array

    cur_offset = 0
    arr_complete = None

    arr = None

    while True:
        # endpoint can be flaky if the response payload is extremely large,
        # so retry at most NUM_RETRIES times
        for i in range(NUM_RETRIES):
            data_args = {**data, "offset": cur_offset}

            out = requests.post(endpoint_url, data=data_args, headers=headers)

            out = json.loads(out.text)

            if out["status"] == 401:
                raise Exception(
                    f"request response is {out} for request {data_args} to endpoint {endpoint_url}, headers {headers}"
                )

            try:
                arr = parse_data(out["body"][arr_key])
                break
            except KeyError:
                if "body" in out.keys():
                    raise Exception(
                        f'got key {arr_key}, expected one of {out["body"].keys()}'
                    )
                elif out["status"] == 2555:
                    # when the payload is too large, this is the status code
                    continue
                else:
                    raise Exception(
                        f"request response is {out} for request {data_args} to endpoint {endpoint_url}, headers {headers}"
                    )

        # for example, https://developer.withings.com/api-reference/#operation/measurev2-getactivity
        # vs. https://developer.withings.com/api-reference/#operation/measure-getmeas

        if arr is not None:
            if type(arr) == type({}):
                if arr_complete is None:
                    arr_complete = dict()

                arr_complete.update(arr)

            elif type(arr) == type([]):
                if arr_complete is None:
                    arr_complete = []

                arr_complete += arr

            # continue if there's still more to get
            if "more" in out["body"].keys() and out["body"]["more"] == 1:
                cur_offset = out["body"]["offset"]
            else:
                break
        else:
            break

    # replace with concatenated version
    if arr_complete is not None:
        out["body"][arr_key] = arr_complete

    return out


def fetch_all_heart_rate(access_token, start="2020-03-10", end="2022-05-28"):
    # get all dates heart rate was collected for
    out = fetch_all_wrapper(
        "https://wbsapi.withings.net/v2/heart",
        {
            "action": "list",
            "startdateymd": start,
            "enddateymd": end,
        },
        {"Authorization": f"Bearer {access_token}"},
        arr_key="series",
    )

    signals = [
        act["ecg"]["signalid"]
        for act in out["body"]["series"]
        if act["ecg"] and act["signalid"]
    ]

    # now for each date get the heart rate data and store as list of dicts
    dict_list = []
    for signal in tqdm(signals):
        out = fetch_all_wrapper(
            "https://wbsapi.withings.net/v2/heart",
            {
                "action": "get",
                "signalid": signal,
            },
            {"Authorization": f"Bearer {access_token}"},
            arr_key="series",
        )

        if "body" in out.keys():
            dict_list += [
                {"datetime": datetime.fromtimestamp(int(k)), **v}
                for k, v in out["body"]["series"].items()
                if datetime.strptime(start, "%Y-%m-%d")
                <= datetime.fromtimestamp(int(k))
                <= datetime.strptime(end, "%Y-%m-%d")
            ]

    df = pd.DataFrame.from_dict(dict_list)

    return df


def fetch_all_sleeps(access_token, start="2020-03-10", end="2022-05-28"):
    start_dt = datetime.strptime(start, "%Y-%m-%d")
    end_dt = datetime.strptime(end, "%Y-%m-%d")

    out = fetch_all_wrapper(
        "https://wbsapi.withings.net/v2/sleep",
        {
            "action": "get",
            "startdate": int(time.mktime(start_dt.timetuple())),
            "enddate": int(time.mktime(end_dt.timetuple())),
            "data_fields": "hr,rr,snoring,sdnn_1,rmssd,mvt_score,chest_movement_rate,withings_index,breathing_sounds",
        },
        {"Authorization": f"Bearer {access_token}"},
        arr_key="series",
    )

    df = pd.DataFrame.from_dict(out["body"]["series"])

    return df


def fetch_all_sleep_summaries(access_token, start="2020-03-10", end="2022-05-28"):
    out = fetch_all_wrapper(
        "https://wbsapi.withings.net/v2/sleep",
        {
            "action": "getsummary",
            "startdateymd": start,
            "enddateymd": end,
            "data_fields": "total_timeinbed,total_sleep_time,asleepduration,lightsleepduration,remsleepduration,deepsleepduration,sleep_efficiency,sleep_latency,wakeup_latency,wakeupduration,wakeupcount,waso,nb_rem_episodes,apnea_hypopnea_index,withings_index,durationtosleep,durationtowakeup,out_of_bed_count,hr_average,hr_min,hr_max,rr_average,rr_min,rr_max,snoring,snoringepisodecount,sleep_score,night_events,mvt_score_avg,mvt_active_duration,chest_movement_rate_average,chest_movement_rate_min,chest_movement_rate_max,breathing_sounds,breathing_sounds_episode_count",
        },
        {"Authorization": f"Bearer {access_token}"},
        arr_key="series",
    )

    df = pd.DataFrame.from_dict(out["body"]["series"])

    return df


#############################
# body+ scale notebook code #
#############################

# @title Extract data via the API!
use_synthetic = True  # @param {type:"boolean"}

num_to_description = {
    1: "Weight (kg)",
    4: "Height (meter)",
    5: "Fat Free Mass (kg)",
    6: "Fat Ratio (%)",
    8: "Fat Mass Weight (kg)",
    9: "Diastolic Blood Pressure (mmHg)",
    10: "Systolic Blood Pressure (mmHg)",
    11: "Heart Pulse (bpm) - only for BPM and scale devices",
    12: "Temperature (celsius)",
    54: "SP02 (%)",
    71: "Body Temperature (celsius)",
    73: "Skin Temperature (celsius)",
    76: "Muscle Mass (kg)",
    77: "Hydration (kg)",
    88: "Bone Mass (kg)",
    91: "Pulse Wave Velocity (m/s)",
    123: "VO2 max is a numerical measurement of your body’s ability to consume oxygen (ml/min/kg).",
    135: "QRS interval duration based on ECG signal",
    136: "PR interval duration based on ECG signal",
    137: "QT interval duration based on ECG signal",
    138: "Corrected QT interval duration based on ECG signal",
    139: "Atrial fibrillation result from PPG",
}


def fetch_measurements(access_token, start, end, measure_types="1,6"):
    # make public API requests, while specifying the measure_types we desire
    # we make potentially multiple because the public API can return only up
    # to 200 measurements

    cur_offset = 0
    data_complete = []
    while True:
        out = requests.post(
            "https://wbsapi.withings.net/measure",
            data={
                "action": "getmeas",
                "meastypes": measure_types,
                "offset": cur_offset,
            },
            headers={"Authorization": f"Bearer {access_token}"},
        )

        # convert this to python dict and get just the actual time series as
        # a pandas dataframe
        out = json.loads(out.text)
        measurements = out["body"]["measuregrps"]

        data = [
            {
                **{"date": datetime.fromtimestamp(meas["date"])},
                **{
                    num_to_description[meas_val["type"]]: meas_val["value"]
                    * 10 ** (meas_val["unit"])
                    for meas_val in meas["measures"]
                },
            }
            for meas in measurements
            if start < datetime.fromtimestamp(meas["date"]) < end
        ]
        # filter by start/end, even though this is kind of mutating what
        # the API gives us raw...

        data_complete += data

        if "more" in out["body"].keys() and out["body"]["more"] == 1:
            cur_offset = out["body"]["offset"]
        else:
            break

    df = pd.DataFrame(data)

    return df



import collections
from datetime import datetime, timedelta

import numpy as np

__all__ = ["create_syn_data"]


def convert_string_to_datetime(date_str):
    date_object = datetime.strptime(date_str, "%Y-%m-%d")
    formatted_date_str = f"{date_object.strftime('%Y-%m-%dT%H:%M:%S.%f')} -00:00"
    return formatted_date_str


def get_daily_activity(date):
    """Generate daily activity data for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :return: daily activity data dictionary
    :rtype: dictionary
    """

    random_number = np.random.randint(100, 999)

    daily_activity = {
        "id": f"fd54d467-4c71-450e-a3ce-7a3951f3d{random_number}",
        "class_5_min": "".join(str(np.random.randint(0, 9)) for _ in range(300)),
        "score": np.random.randint(50, 100),
        "active_calories": np.random.randint(300, 1500),
        "average_met_minutes": np.random.randint(500, 2000) / 1000,
        "contributors": {
            "meet_daily_targets": np.random.randint(50, 100),
            "move_every_hour": np.random.randint(50, 100),
            "recovery_time": np.random.randint(50, 100),
            "stay_active": np.random.randint(50, 100),
            "training_frequency": np.random.randint(50, 100),
            "training_volume": np.random.randint(50, 100),
        },
        "equivalent_walking_distance": np.random.randint(500, 10000),
        "high_activity_met_minutes": np.random.randint(500, 2000) / 1000,
        "high_activity_time": np.random.randint(500, 2000) / 1000 * 60,
        "inactivity_alerts": np.random.randint(0, 10),
        "low_activity_met_minutes": np.random.randint(500, 2000) / 1000 * 60,
        "low_activity_time": np.random.randint(500, 2000) / 1000 * 60,
        "medium_activity_met_minutes": np.random.randint(500, 2000) / 1000 * 60,
        "medium_activity_time": np.random.randint(500, 2000) / 1000 * 60,
        "met": {
            "interval": 60.0,
            "timestamp": convert_string_to_datetime(date),
            "items": [round(np.random.uniform(0.9, 5.0), 1) for _ in range(300)],
        },
        "meters_to_target": np.random.randint(50, 2000),
        "non_wear_time": np.random.randint(50, 18000),
        "resting_time": np.random.randint(50, 18000),
        "sedentary_met_minutes": np.random.randint(500, 2000) / 1000 * 60,
        "sedentary_time": np.random.randint(500, 2000) / 1000 * 60,
        "steps": np.random.randint(500, 10000),
        "target_calories": 500,
        "target_meters": 10000,
        "total_calories": np.random.randint(1500, 4000),
        "day": date,
        "timestamp": convert_string_to_datetime(date),
    }
    return daily_activity


def get_daily_sleep(date):
    daily_sleep_data = {
        "id": 1,
        "contributors": {
            "deep_sleep": np.random.randint(1, 100),
            "efficiency": np.random.randint(1, 100),
            "latency": np.random.randint(1, 100),
            "rem_sleep": np.random.randint(1, 100),
            "restfulness": np.random.randint(1, 100),
            "timing": np.random.randint(1, 100),
            "total_sleep": np.random.randint(1, 100),
        },
        "day": date,
        "score": np.random.randint(1, 100),
        "timestamp": convert_string_to_datetime(date),
    }
    return daily_sleep_data


def get_sleep(date):
    """Generate sleep data for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :return: sleep data dictionary
    :rtype: dictionary
    """
    summary_date = datetime.strptime(date, "%Y-%m-%d")

    period_id = 1
    is_longest = 1
    timezone = np.random.randint(-600, 600)
    bedtime_end = (
        summary_date + timedelta(minutes=np.random.randint(0, 1440))
    ).isoformat()
    bedtime_start = (
        summary_date + timedelta(minutes=np.random.randint(0, 1440))
    ).isoformat()
    sleep_type = np.random.choice(["long_sleep", "short_sleep"])
    breath_average = round(np.random.uniform(10, 20), 3)
    average_breath_variation = round(np.random.uniform(0, 5), 3)
    duration = np.random.randint(18000, 43200)
    total = np.random.randint(int(duration * 0.9), duration)
    awake = np.random.randint(0, int(duration * 0.1))
    rem = np.random.randint(int(duration * 0.2), int(duration * 0.3))
    deep = np.random.randint(int(duration * 0.4), int(duration * 0.6))
    light = total - rem - deep
    midpoint_time = np.random.randint(int(duration * 0.4), int(duration * 0.6))
    efficiency = np.random.randint(80, 100)
    restless = np.random.randint(5 * 60, 30 * 60)
    onset_latency = np.random.randint(600, 1800)
    got_up_count = 0
    wake_up_count = np.random.randint(0, 10)
    hr_5min = [np.random.randint(50, 60) for _ in range(96)] + [0]

    hr_average = round(sum(hr_5min) / len(hr_5min), 3)
    hr_lowest = min(hr_5min)
    lowest_heart_rate_time_offset = hr_5min.index(hr_lowest) * 300

    hypnogram_5min = "".join([str(np.random.randint(1, 4)) for _ in range(96)])

    rmssd_5min = [np.random.randint(20, 120) for _ in range(96)] + [0]
    rmssd = round(sum(rmssd_5min) / len(rmssd_5min))

    score = np.random.randint(60, 90)
    score_alignment = np.random.randint(40, 80)
    score_deep = np.random.randint(80, 100)
    score_disturbances = np.random.randint(60, 90)
    score_efficiency = np.random.randint(80, 100)
    score_latency = np.random.randint(60, 90)
    score_rem = np.random.randint(40, 70)
    score_total = np.random.randint(50, 80)

    temperature_deviation = round(np.random.uniform(-1, 1), 2)
    temperature_trend_deviation = round(np.random.uniform(-0.1, 0.1), 2)
    bedtime_start_delta = np.random.randint(3000, 4000)
    bedtime_end_delta = np.random.randint(18000, 43200)
    midpoint_at_delta = np.random.randint(7200, 21600)
    temperature_delta = round(np.random.uniform(-1, 1), 2)

    sleep_data = {
        "id": period_id,
        "average_breath": breath_average,
        "average_heart_rate": hr_average,
        "average_hrv": np.random.randint(45, 65),
        "awake_time": awake,
        "bedtime_end": bedtime_end,
        "bedtime_start": bedtime_start,
        "day": summary_date.strftime("%Y-%m-%d"),
        "deep_sleep_duration": deep,
        "efficiency": efficiency,
        "heart_rate": {
            "interval": 0,
            "items": [0],
            "timestamp": convert_string_to_datetime(date),
        },
        "hrv": {
            "interval": 0,
            "items": [0],
            "timestamp": convert_string_to_datetime(date),
        },
        "latency": np.random.randint(600, 1800),
        "light_sleep_duration": light,
        "low_battery_alert": False,
        "lowest_heart_rate": hr_lowest,
        "movement_30_sec": "1",
        "period": period_id,
        "readiness": {
            "contributors": {
                "activity_balance": np.random.randint(1, 100),
                "body_temperature": np.random.randint(1, 100),
                "hrv_balance": np.random.randint(1, 100),
                "previous_day_activity": np.random.randint(1, 100),
                "previous_night": np.random.randint(1, 100),
                "recovery_index": np.random.randint(1, 100),
                "resting_heart_rate": np.random.randint(1, 100),
                "sleep_balance": np.random.randint(1, 100),
            },
            "score": np.random.randint(1, 100),
            "temperature_deviation": round(np.random.uniform(-1, 1), 2),
            "temperature_trend_deviation": round(np.random.uniform(-1, 1), 2),
        },
        "readiness_score_delta": np.random.randint(-100, 100),
        "rem_sleep_duration": rem,
        "restless_periods": restless,
        "sleep_phase_5_min": "1",
        "sleep_score_delta": np.random.randint(-100, 100),
        "sleep_algorithm_version": "1.0.0",
        "time_in_bed": total + awake,
        "total_sleep_duration": total,
        "type": sleep_type,
    }
    return sleep_data


def get_activity(date):
    """Generate activity data for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :return: dictionary of activity data
    :rtype: dictionary
    """
    summary_date = datetime.strptime(date, "%Y-%m-%d")
    start_time = summary_date.replace(hour=4, minute=0, second=0)
    end_time = start_time + timedelta(days=1) - timedelta(seconds=1)
    timezone_offset = -420

    cal_active = np.random.randint(300, 800)
    cal_total = cal_active + np.random.randint(1500, 2000)
    steps = np.random.randint(8000, 15000)
    daily_movement = np.random.randint(7000, 12000)

    non_wear = np.random.randint(0, 120) * 60
    rest = np.random.randint(400, 600) * 60
    inactive = np.random.randint(500, 800)
    low = np.random.randint(200, 400)
    medium = np.random.randint(50, 150) * 60
    high = np.random.randint(0, 20)
    inactivity_alerts = np.random.randint(0, 3)
    average_met = round(np.random.uniform(1.0, 1.8), 2)

    met_1min = [round(np.random.uniform(0.9, 5.0), 1) for _ in range(1440)]

    activity_data = {
        "summary_date": summary_date.strftime("%Y-%m-%d"),
        "timezone": timezone_offset,
        "day_start": start_time.isoformat(),
        "day_end": end_time.isoformat(),
        "cal_active": cal_active,
        "cal_total": cal_total,
        "class_5min": "".join(str(np.random.randint(0, 5)) for _ in range(288)),
        "steps": steps,
        "daily_movement": daily_movement,
        "non_wear": non_wear,
        "rest": rest,
        "inactive": inactive,
        "low": low,
        "medium": medium,
        "high": high,
        "inactivity_alerts": inactivity_alerts,
        "average_met": average_met,
        "met_1min": met_1min,
    }

    return activity_data


def get_readiness(date):
    """Generate readiness data for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :return: daily activity data dictionary
    :rtype: dictionary
    """

    summary_date = datetime.strptime(date, "%Y-%m-%d")
    score = np.random.randint(70, 100)
    score_activity_balance = np.random.randint(70, 100)
    score_hrv_balance = np.random.randint(70, 100)
    score_previous_day = np.random.randint(70, 100)
    score_previous_night = np.random.randint(70, 100)
    score_recovery_index = np.random.randint(70, 100)
    score_resting_hr = np.random.randint(70, 100)
    score_sleep_balance = np.random.randint(70, 100)
    score_temperature = np.random.randint(70, 99)

    rest_mode_state = np.random.randint(0, 1)
    period_id = np.random.randint(1, 4)

    readiness_data = {
        "summary_date": summary_date.strftime("%Y-%m-%d"),
        "score": score,
        "score_activity_balance": score_activity_balance,
        "score_hrv_balance": score_hrv_balance,
        "score_previous_day": score_previous_day,
        "score_previous_night": score_previous_night,
        "score_recovery_index": score_recovery_index,
        "score_resting_hr": score_resting_hr,
        "score_sleep_balance": score_sleep_balance,
        "score_temperature": score_temperature,
        "rest_mode_state": rest_mode_state,
        "period_id": period_id,
    }

    return readiness_data


def get_ideal_bedtime(date):
    """Generate ideal bedtime data for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :return: daily activity data dictionary
    :rtype: dictionary
    """
    date = datetime.strptime(date, "%Y-%m-%d")

    start_time = np.random.randint(0, 86400)
    end_time = np.random.randint(start_time, 86400)

    ideal_bedtime_data = {
        "date": date.strftime("%Y-%m-%d"),
        "bedtime_window": {"start": start_time, "end": end_time},
        "status": "IDEAL_BEDTIME_AVAILABLE",
    }

    return ideal_bedtime_data


def get_heart_rate(date):
    """Generate heart rate data for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :return: heart rate data dictionary
    :rtype: dictionary
    """
    date = datetime.strptime(date, "%Y-%m-%d")

    timestamp = date.timestamp()

    heart_rate_data = []

    number_of_minues_in_a_day = 1440

    bpm = np.random.randint(50, 120)

    for i in range(0, number_of_minues_in_a_day, 5):
        # Calculate the current hour based on the timestamp
        hour = int((date + timedelta(minutes=i)).strftime("%H"))

        # Determine if the person is awake (6 AM to 10 PM)
        is_awake = 6 <= hour < 22
        status = "awake" if is_awake else "asleep"

        # Gradually change the heart rate within a realistic range
        if is_awake:
            # Simulate an increase in heart rate during awake hours
            bpm += np.random.randint(-3, 5)
            bpm = min(
                np.random.randint(100, 120), bpm
            )  # Ensure heart rate doesn't exceed 120 bpm
        else:
            # Simulate a decrease in heart rate during asleep hours
            bpm -= np.random.randint(-3, 5)
            bpm = max(
                np.random.randint(50, 70), bpm
            )  # Ensure heart rate doesn't go below 50 bpm

        formatted_timestamp = (date + timedelta(minutes=i)).strftime(
            "%Y-%m-%dT%H:%M:%S+00:00"
        )

        heart_rate_entry = {
            "bpm": bpm,
            "source": status,
            "timestamp": formatted_timestamp,
        }

        heart_rate_data.append(heart_rate_entry)

    return heart_rate_data


def get_session(date):
    return []


def get_enhanced_tag(date):
    return []


def get_workout(date):
    return []


def create_syn_data(seed, start_date, end_date):
    """Returns a dict of daily activity data, sleep data, ideal bedtime, readiness, and activity

    :param start_date: the start date (inclusive) as a string in the format "YYYY-MM-DD"
    :type start_date: str
    :param end_date: the end date (inclusive) as a string in the format "YYYY-MM-DD"
    :type end_date: str
    :return: a defaultdict of daily activity data, sleep data, ideal bedtime, readiness, and activity
    :rtype: defaultdict
    """
    np.random.seed(seed)

    num_days = (
        datetime.strptime(end_date, "%Y-%m-%d")
        - datetime.strptime(start_date, "%Y-%m-%d")
    ).days

    # first get the dates as datetime objects
    synth_dates = [
        datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=i)
        for i in range(num_days)
    ]

    for i, date in enumerate(synth_dates):
        synth_dates[i] = date.strftime("%Y-%m-%d")

    full_dict = collections.defaultdict(list)

    for date in synth_dates:
        full_dict["heart_rate"].extend(get_heart_rate(date))
        full_dict["session"].extend(get_session(date))
        full_dict["enhanced_tag"].extend(get_enhanced_tag(date))
        full_dict["workout"].extend(get_workout(date))
        full_dict["daily_activity"].append(get_daily_activity(date))
        full_dict["daily_sleep"].append(get_daily_sleep(date))
        full_dict["sleep"].append(get_sleep(date))
        full_dict["readiness"].append(get_readiness(date))
        full_dict["ideal_sleep_time"].append(get_ideal_bedtime(date))

    return full_dict



import numpy as np
import pandas as pd
import polyline

# Generating synthetic data for Strava


def create_syn_data(start_date, end_date):
    """
    This function generates synthetic data for Strava.

    :param start_date: The start date for the data generation in the format "YYYY-MM-DD".
    :type start_date: str
    :param end_date: The end date for the data generation in the format "YYYY-MM-DD".
    :type end_date: str
    :return: A dataframe containing the generated data.
    :rtype: pandas.DataFrame
    """

    # Generating list of dates between start and end date
    dates = pd.date_range(start_date, end_date)

    # Generating name for run
    def name_generator(x):
        return (
            np.random.choice(
                [
                    "Morning",
                    "Post Lunch",
                    "Evening",
                    "Night",
                    "Late Night",
                    "Post Dinner",
                ]
            )
            + " Run"
        )

    # Generating 10 digit random id
    def id_generator(x):
        return np.random.randint(1000000000, 9999999999, dtype=np.int64)

    # Generating random distance with average run being 3000m and standard deviation of 750m
    def distance_generator(x):
        return np.round(np.random.normal(3000, 750), 1)

    # Generating random moving time with average run being 1200s and standard deviation of 300s
    def moving_time_generator(x):
        return int(np.random.normal(1200, 300))

    # Generating random elapsed time with average run being 1700s and standard deviation of 300s
    def elapsed_time_generator(x):
        return int(np.random.normal(1700, 300))

    # Generating random total elevation gain with average run being 20m and standard deviation of 30m
    def total_elevation_gain_generator(x):
        return max(0.0, np.round(np.random.normal(20, 30), 1))

    # Generating random average speed with average run being 2.5m/s and standard deviation of 1.5m/s
    def average_speed_generator(x):
        return max(0.0, np.round(np.random.normal(2.5, 1.5, 1)))

    # Generating random max speed with average run being 3.5m/s and standard deviation of 1.5m/s
    def max_speed_generator(x):
        return max(0.0, np.round(np.random.normal(3.5, 1.5, 1)))

    # Generating random average heartrate with average run being 130bpm and standard deviation of 20bpm
    def average_heartrate_generator(x):
        return np.round(np.random.normal(130, 20), 1)

    # Generating random max heartrate with average run being 155bpm and standard deviation of 20bpm
    def max_heartrate_generator(x):
        return np.round(np.random.normal(155, 20), 1)

    # We will use Stanford's coordinate as the benchmark coordinates to generate random runs based off
    coordinate = (37.4275, -122.1697)

    # Generating random elev_high with average run being 40m and standard deviation of 25m
    def elev_high_generator(x):
        return max(0.0, np.round(np.random.normal(40, 25), 1))

    # Generating random elev_low with average run being 20m and standard deviation of 20m
    def elev_low_generator(x):
        return max(0.0, np.round(np.random.normal(20, 20), 1))

    # Generating random average cadence with average run being 75rpm and standard deviation of 10rpm
    def avg_cadence_generator(x):
        return np.round(np.random.normal(75, 10), 1)

    # Generating random average watts with average run being 190w and standard deviation of 50w
    def avg_watt_generator(x):
        return np.round(np.random.normal(190, 50), 1)

    # Generating random kilojoules with average run being 0.0864kj/m
    def kilo_joule_gen(distance):
        return np.round(distance * 0.0864, 1)

    # Generating random map summary polyline
    def map_summary_polyline_generator(x):
        return polyline.encode(x)

    # Creating a dataframe to store all the generated data
    synthetic_df = pd.DataFrame(
        columns=[
            "name",
            "id",
            "start_date",
            "distance",
            "moving_time",
            "elapsed_time",
            "total_elevation_gain",
            "average_speed",
            "max_speed",
            "average_heartrate",
            "max_heartrate",
            "map.summary_polyline",
            "elev_high",
            "elev_low",
            "average_cadence",
            "average_watts",
            "kilojoules",
        ]
    )

    for d in dates:

        # Generating 12 different coordinates for each run
        coordinates = [
            (
                coordinate[0] + np.random.normal(-0.005, 0.005),
                coordinate[1] + np.random.normal(-0.005, 0.005),
            )
            for x in range(12)
        ]

        # Generating random run distance
        distance = distance_generator(d)

        # Generating entry of all synthetic data to be added to the dataframe
        entry = pd.DataFrame.from_dict(
            {
                "name": name_generator(d),
                "id": id_generator(d),
                "start_date": d.strftime("%Y-%m-%dT%H:%M:%SZ"),
                "distance": distance,
                "moving_time": moving_time_generator(d),
                "elapsed_time": elapsed_time_generator(d),
                "total_elevation_gain": total_elevation_gain_generator(d),
                "average_speed": average_speed_generator(d),
                "max_speed": max_speed_generator(d),
                "average_heartrate": average_heartrate_generator(d),
                "max_heartrate": max_heartrate_generator(d),
                "map.summary_polyline": map_summary_polyline_generator(coordinates),
                "elev_high": elev_high_generator(d),
                "elev_low": elev_low_generator(d),
                "average_cadence": avg_cadence_generator(d),
                "average_watts": avg_watt_generator(d),
                "kilojoules": kilo_joule_gen(distance),
            }
        )
        synthetic_df = pd.concat([synthetic_df, entry], ignore_index=True)

    # Returning the synthetic dataframe
    return synthetic_df



from QualtricsAPI.Setup import Credentials
from QualtricsAPI.Survey import Responses


def fetch_real_data(survey):
    """
    Fetches real survey data from Qualtrics.
    This function uses QualtricsAPI to fetch survey responses based on the given survey ID.
    The responses are retrieved with labels (actual user responses) instead of numeric codes.

    :param survey: the ID of the survey from which to fetch responses
    :type start_date: str

    :return: the survey responses fetched from the API
    :rtype: Pandas DataFrame
    """
    responses = Responses().get_survey_responses(survey=survey, useLabels=True)
    return responses



import requests


def fetch_real_data(start_date, end_date, data_type, headers):
    """Main function for fetching real data from the nutrisense database.
    Uses Nutrisense's internal API.

    :param start_date: the start date represented as a string in the format "YYYY-MM-DD"
    :type start_date: str
    :param end_date: the end date represented as a string in the format "YYYY-MM-DD"
    :type end_date: str
    :param data_type: the type of data to fetch, one of "continiuous", "summary", "scores", "statistics"
    :type data_type: str
    :param headers: current header with credentials to Nutrisense, pre authenticated
    :type headers: requests.sessions.Session
    :return: the data fetched from the API according to the inputs
    :rtype: list[dict] (for continuous data) or dict (otherwise)
    """

    if data_type == "continuous" or data_type == "summary":
        json_data = {
            "operationName": "allCharts",
            "variables": {
                "filter": {
                    "types": [
                        {
                            "key": "timeline",
                            "value": [],
                        },
                    ],
                    "startDate": start_date,
                    "endDate": end_date,
                },
            },
            "query": "query allCharts($filter: DateFilter) {\n  allCharts(filter: $filter) {\n    charts {\n      type\n      title\n      description\n      xAxis\n      yAxis\n      range {\n        min\n        max\n        goal\n        goalMin\n        goalMax\n        __typename\n      }\n      meta {\n        key\n        tag\n        section\n        __typename\n      }\n      values {\n        ... on TimePair {\n          x\n          y\n          interpolated\n          __typename\n        }\n        ... on NumericPair {\n          x\n          y\n          __typename\n        }\n        ... on StringPair {\n          name\n          x\n          y\n          __typename\n        }\n        ... on RangePair {\n          x {\n            min\n            max\n            __typename\n          }\n          y\n          __typename\n        }\n        __typename\n      }\n      __typename\n    }\n    __typename\n  }\n}",
        }

        response = requests.post(
            "https://api-production.nutrisense.io/graphql",
            headers=headers,
            json=json_data,
            verify=False,
        )
        res = response.json()

        if data_type == "continuous":
            return res["data"]["allCharts"]["charts"][0]["values"]
        elif data_type == "summary":
            return res["data"]["allCharts"]["charts"][0]["range"]

    else:
        json_data = {
            "operationName": "allNutrition",
            "variables": {
                "filter": {
                    "startDate": start_date,
                    "endDate": end_date,
                },
            },
            "query": "query allNutrition($filter: DateFilter) {\n  allNutrition(filter: $filter) {\n    nutrition {\n      today {\n        key\n        value\n        __typename\n      }\n      average {\n        key\n        value\n        __typename\n      }\n      __typename\n    }\n    score {\n      today {\n        scoreTimeOutsideRange\n        scorePeak\n        scoreMean\n        scoreStdDev\n        score\n        __typename\n      }\n      __typename\n    }\n    statistics {\n      today {\n        healthyRange {\n          min\n          max\n          __typename\n        }\n        range {\n          min\n          max\n          __typename\n        }\n        timeWithinRange\n        min\n        max\n        mean\n        median\n        standardDeviation\n        q1\n        q3\n        score\n        __typename\n      }\n      average {\n        healthyRange {\n          min\n          max\n          __typename\n        }\n        range {\n          min\n          max\n          __typename\n        }\n        timeWithinRange\n        min\n        max\n        mean\n        median\n        standardDeviation\n        q1\n        q3\n        score\n        __typename\n      }\n      __typename\n    }\n    __typename\n  }\n}",
        }

        response = requests.post(
            "https://api-production.nutrisense.io/graphql",
            headers=headers,
            json=json_data,
            verify=False,
        )
        res = response.json()

        if data_type == "scores":
            return res["data"]["allNutrition"]["score"]["today"]
        elif data_type == "statistics":
            return {
                "today": res["data"]["allNutrition"]["statistics"]["today"],
                "average": res["data"]["allNutrition"]["statistics"]["average"],
            }



import collections
import random
from datetime import datetime, timedelta
from random import choice, randrange

import numpy as np

__all__ = ["create_syn_data"]

sleep_stages = {
    "deepSleepSummary": {"mean": 12, "std": 1.5},
    "remSleepSummary": {"mean": 16, "std": 1.5},
    "lightSleepSummary": {"mean": 18, "std": 2},
}


def get_sleep(date):
    """Generate sleep data for a given date.


    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :return: sleep data dictionary
    :rtype: dictionary
    """

    duration = random.randint(14400, 36000)

    awake = random.randint(2, 9)
    afterwake = random.randint(0, 200) / 100
    tofall = random.randint(0, 200) / 100

    percents = (100 - awake - afterwake - tofall, awake, afterwake, tofall)
    start = datetime.strptime(date + " 9:00 PM", "%Y-%m-%d %I:%M %p")
    end = datetime.strptime(date + " 11:58 PM", "%Y-%m-%d %I:%M %p")

    # generate random date
    delta = end - start
    int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
    random_second = randrange(int_delta)

    start_time = start + timedelta(seconds=random_second)
    end_time = start_time + timedelta(seconds=duration)

    sleep_dict = {
        "dateOfSleep": date,
        "duration": duration * 1000,
        "efficiency": random.randint(90, 99),
        "endTime": str(end_time.isoformat()),
        "infoCode": 0,
        "isMainSleep": True,
        "logId": random.randint(0, 1000000000),
        "logType": "auto_detected",
        "minutesAfterWakeup": round(duration / 60 * percents[2] / 100),
        "minutesAsleep": round(duration / 60 * percents[0] / 100),
        "minutesAwake": round(duration / 60 * percents[1] / 100),
        "minutesToFallAsleep": round(duration / 60 * percents[3] / 100),
        "startTime": str(start_time.isoformat()),
        "timeInBed": round(duration / 60),
        "levels": dict(),
        "type": "stages",
    }
    sleep_dict["levels"]["data"] = []
    sleep_dict["levels"]["shortData"] = []
    sleep_dict["levels"]["summary"] = {
        "deep": {"count": 0, "minutes": 0},
        "light": {"count": 0, "minutes": 0},
        "rem": {"count": 0, "minutes": 0},
        "wake": {"count": 2, "minutes": round(duration / 60 * percents[1] / 100)},
    }
    #  core sleep levels
    sleep_dict["levels"]["data"].append(
        {
            "dateTime": str(start_time.isoformat()),
            "level": "wake",
            "seconds": round(duration / 60 * percents[3] / 100) * 60,
        }
    )
    # split sleep times
    def split_the_duration(duration):
        saver_of_duration = duration
        while duration > 1:
            n = random.randint(1, round(saver_of_duration / 12))
            if duration - n >= 0:
                yield n
            else:
                yield duration
            duration -= n

    generator = split_the_duration(round(duration * percents[0] / 100))

    arr_of_durations = list(generator)

    start_phases = start_time + timedelta(
        seconds=round(duration / 60 * percents[3] / 100) * 60
    )

    start_phases += timedelta(seconds=arr_of_durations[-1])
    sleep_dict["levels"]["data"].append(
        {
            "dateTime": str(start_phases.isoformat()),
            "level": "wake",
            "seconds": round(duration / 60 * percents[2] / 100) * 60,
        }
    )

    # do the same for wakeups
    generator = split_the_duration(round(duration * percents[0] / 100))
    arr_of_durations = list(generator)

    for i, item in enumerate(arr_of_durations):

        if i != 0:
            start_phases += timedelta(seconds=arr_of_durations[i - 1])

        type = choice(["deep", "rem", "light"])

        sleep_dict["levels"]["summary"][type]["count"] += 1
        sleep_dict["levels"]["summary"][type]["minutes"] += round(item / 60)

        sleep_dict["levels"]["data"].append(
            {"dateTime": str(start_phases.isoformat()), "level": type, "seconds": item}
        )

        if i != 0:
            start_phases += timedelta(seconds=arr_of_durations[i - 1])
            sleep_dict["levels"]["summary"]["wake"]["count"] += 1
        # generate random date
        delta = end_time - start_time
        int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
        random_second = randrange(int_delta)
        randtime = start_time + timedelta(seconds=random_second)
        start_time = start_time + timedelta(seconds=arr_of_durations[i - 1])

        sleep_dict["levels"]["shortData"].append(
            {"dateTime": str(randtime.isoformat()), "level": "wake", "seconds": item}
        )

    return sleep_dict


def get_activity(date):
    """Generate activity data for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :return: dictionaries of "steps", "minutesVeryActive", "minutesFairlyActive", "minutesLightlyActive", "distance", "minutesSedentary"
    :rtype: dictionary
    """

    very_active = random.randint(0, 240)
    fairly_active = random.randint(0, 240)
    lightly_active = random.randint(0, 240)

    minutes_in_a_day = 1440

    steps = {
        "dateTime": date,
        "value": very_active * 50 + fairly_active * 25 + lightly_active * 10,
    }
    minutesVeryActive = {"dateTime": date, "value": very_active}
    minutesFairlyActive = {"dateTime": date, "value": fairly_active}
    minutesLightlyActive = {"dateTime": date, "value": lightly_active}
    distance = {
        "dateTime": date,
        "value": (very_active * 50 + fairly_active * 25 + lightly_active * 10) / 1326,
    }
    minutesSedentary = {
        "dateTime": date,
        "value": minutes_in_a_day - (very_active + fairly_active + lightly_active),
    }

    return (
        steps,
        minutesVeryActive,
        minutesFairlyActive,
        minutesLightlyActive,
        distance,
        minutesSedentary,
    )


def get_heart_rate(date, intraday=False):
    """Generate heart rate data for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :param intraday: whether the heart rate is reported per second
    :type intraday: bool
    :return: dictionary of heart rate values and details
    :rtype: dictionary
    """

    HR_MEAN = 75
    HR_STD = 15
    HR_MIN = 50
    HR_MAX = 195
    ZONE_CALORIES = [0, 1, 2, 3]

    heart_rate_zones = [
        {
            "caloriesOut": 2877.06579,
            "max": 110,
            "min": 30,
            "minutes": 0,
            "name": "Out of Range",
        },
        {"caloriesOut": 1000, "max": 136, "min": 110, "minutes": 0, "name": "Fat Burn"},
        {"caloriesOut": 1000, "max": 169, "min": 136, "minutes": 0, "name": "Cardio"},
        {"caloriesOut": 1000, "max": 220, "min": 169, "minutes": 0, "name": "Peak"},
    ]

    the_time = datetime.strptime(date, "%Y-%m-%d").replace(hour=0, minute=0, second=0)
    iterations_in_a_day = 1440 if not intraday else 1440 * 60

    bpm_start = np.clip(np.random.normal(HR_MEAN, HR_STD), HR_MIN, HR_MAX)
    time_intervals = np.arange(iterations_in_a_day)

    random_walk = np.random.randint(-1, 2, size=iterations_in_a_day)
    heart_rate_values = np.clip(np.cumsum(random_walk) + bpm_start, HR_MIN, HR_MAX)

    hours = (time_intervals // 60) % 24

    zone_indices = np.where(
        hours < 6, 0, np.where(hours < 10, 1, np.where(hours < 18, 2, 3))
    )

    zone_minutes = np.bincount(zone_indices, minlength=4)
    zone_calories = np.array([ZONE_CALORIES[i] * zone_minutes[i] for i in range(4)])

    for i, zone in enumerate(heart_rate_zones):
        zone["minutes"] += zone_minutes[i]
        zone["caloriesOut"] += zone_calories[i]

    dataset = [
        {
            "time": (
                the_time + timedelta(seconds=(i if intraday else i * 60))
            ).strftime("%H:%M:%S"),
            "value": bpm,
        }
        for i, bpm in enumerate(heart_rate_values)
    ]

    heart_rate_data = {
        "heart_rate_day": [
            {
                "activities-heart": [
                    {
                        "dateTime": date,
                        "value": {
                            "customHeartRateZones": [],
                            "heartRateZones": heart_rate_zones,
                            "restingHeartRate": 58,
                        },
                    }
                ],
                "activities-heart-intraday": {
                    "dataset": dataset,
                    "datasetInterval": 1,
                    "datasetType": "minute" if not intraday else "second",
                },
            }
        ]
    }

    return heart_rate_data


def get_intraday_azm(date, hr):
    """Generate active zone minutes for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :param hr: heart rate data collected per second on the same day as date
    :type hr: dict
    :return: dictionary of intraday active zone minute details
    :rtype: dictionary
    """
    azm = {
        "activities-active-zone-minutes-intraday": [{"dateTime": date, "minutes": []}]
    }

    the_time = datetime.strptime(date, "%Y-%m-%d").replace(hour=0, minute=0, second=0)

    minutes_in_a_day = 1440
    hr_dataset = hr["heart_rate_day"][0]["activities-heart-intraday"]["dataset"]
    dataset_length = len(hr_dataset)

    mean_hr_per_minute = [
        np.mean(
            [
                hr_dataset[j]["value"]
                for j in range(i * 60, min((i + 1) * 60, dataset_length))
            ]
        )
        if i * 60 < dataset_length
        else hr_dataset[-1]["value"]
        for i in range(minutes_in_a_day)
    ]

    def get_activity_value(hr):
        if hr > 111:
            return {"peakActiveZoneMinutes": 1, "activeZoneMinutes": 1}
        elif hr > 98:
            return {"cardioActiveZoneMinutes": 1, "activeZoneMinutes": 1}
        elif hr > 87:
            return {"fatBurnActiveZoneMinutes": 1, "activeZoneMinutes": 1}
        else:
            return {"activeZoneMinutes": 0}

    azm["activities-active-zone-minutes-intraday"][0]["minutes"] = [
        {
            "minute": (the_time + timedelta(minutes=i)).strftime("%H:%M:%S"),
            "value": get_activity_value(mean_hr_per_minute[i]),
        }
        for i in range(minutes_in_a_day)
    ]

    return azm


def get_intraday_breath_rate(date):
    """Generate breath rate during sleep for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :return: dictionary of breath rate during sleep details
    :rtype: dictionary
    """
    breathing_rates = {
        stage: np.random.normal(info["mean"], info["std"])
        for stage, info in sleep_stages.items()
    }
    full_breathing_rate = np.mean(list(breathing_rates.values()))

    br = {
        "br": [
            {
                "value": {
                    **{
                        stage: {"breathingRate": rate}
                        for stage, rate in breathing_rates.items()
                    },
                    "fullSleepSummary": {"breathingRate": full_breathing_rate},
                },
                "dateTime": date,
            }
        ]
    }

    return br


def get_hrv(date):
    """Generate hrv for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :return: dictionary of hrv details
    :rtype: dictionary
    """
    hrv = {
        "hrv": [
            {
                "hrv": [
                    {
                        "value": {
                            "dailyRmssd": random.randint(13, 48),
                            "deepRmssd": random.randint(13, 48),
                        },
                        "dateTime": date,
                    }
                ]
            }
        ]
    }

    return hrv


def get_random_sleep_start_time():
    """Generate a random start time for sleep in terms of hour, minute and second, and a random duration of sleep in minutes."""
    random_hour = random.choice([21, 22, 23, 0, 1, 2])
    random_min = random.randint(0, 59)
    random_sec = random.randint(0, 59)
    random_duration = random.randint(360, 540)

    return random_hour, random_min, random_sec, random_duration


def get_intraday_hrv(date, random_hour, random_min, random_sec, random_duration):
    """Generate intraday HRV data for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :param random_hour: the starting hour
    :type random_hour: int
    :param random_min: the starting minute
    :type random_min: int
    :param random_sec: the starting second
    :type random_sec: int
    :param random_duration: the number of minutes for which to generate data
    :type random_duration: int
    :return: dictionary of intraday HRV data details
    :rtype: dictionary
    """

    hrv = {"hrv": [{"minutes": [], "dateTime": date}]}

    the_time = datetime.strptime(date, "%Y-%m-%d").replace(
        hour=random_hour, minute=random_min, second=random_sec
    )

    random_values = np.random.uniform(0, 1, (random_duration, 4))
    hf_values = np.round(100 + 900 * random_values[:, 0], 3)
    rmssd_values = np.round(20 + 60 * random_values[:, 1], 3)
    coverage_values = np.round(0.9 + 0.09 * random_values[:, 2], 3)
    lf_values = np.round(hf_values * (0.2 + 0.2 * random_values[:, 3]), 3)

    times = [the_time + timedelta(minutes=i) for i in range(random_duration)]

    hrv["hrv"][0]["minutes"] = [
        {
            "minute": f"{date}T{time.strftime('%H:%M:%S')}.000",
            "value": {
                "rmssd": rmssd,
                "coverage": coverage,
                "hf": hf,
                "lf": lf,
            },
        }
        for hf, rmssd, coverage, lf, time in zip(
            hf_values, rmssd_values, coverage_values, lf_values, times
        )
    ]

    return hrv


def get_intraday_spo2(date, random_hour, random_min, random_sec, random_duration):
    """Generate intraday SpO2 data for a given date.

    :param date: the date as a string in the format "YYYY-MM-DD"
    :type date: str
    :param random_hour: the starting hour
    :type random_hour: int
    :param random_min: the starting minute
    :type random_min: int
    :param random_sec: the starting second
    :type random_sec: int
    :param random_duration: the number of minutes for which to generate data
    :type random_duration: int
    :return: dictionary of intraday SpO2 data details
    :rtype: dictionary
    """

    spo2_data = {"dateTime": date, "minutes": []}

    the_time = datetime.strptime(date, "%Y-%m-%d").replace(
        hour=random_hour, minute=random_min, second=random_sec
    )

    mean = 97.5
    std_dev = 3
    spo2_value = round(np.random.normal(mean, std_dev), 1)
    spo2_value = max(95, min(100, spo2_value))

    random_changes = np.round(np.random.uniform(-0.5, 0.5, random_duration), 1)
    times = [the_time + timedelta(minutes=i) for i in range(random_duration)]

    spo2_data["minutes"] = [
        {
            "value": (spo2_value := max(95, min(100, spo2_value + change))),
            "minute": time.strftime("%Y-%m-%dT%H:%M:%S"),
        }
        for change, time in zip(random_changes, times)
    ]

    return spo2_data


def get_distance_day(date):
    """Generate distance data for a given date.

    :return: dictionary of distance data details
    :rtype: dictionary
    """

    distance_day = {
        "distance_day": [
            {
                "activities-distance": [{"dateTime": date, "value": 0}],
                "activities-distance-intraday": {
                    "dataset": [],
                    "datasetInterval": 1,
                    "datasetType": "minute",
                },
            }
        ]
    }

    minutes_in_a_day = 1440
    distance = [0, 0.1]
    weights = [0.6, 0.3]

    the_time = datetime.strptime(date, "%Y-%m-%d").replace(hour=0, minute=0, second=0)

    for i in range(minutes_in_a_day):
        distance = [0, 0.1]
        weights = [0.6, 0.3]
        max_distance = random.choices(distance, weights)
        if max_distance[0] == 0:
            val = 0
        else:
            val = random.randint(1, 1000) / 10000

        if 0 <= the_time.hour < 6 or 21 <= the_time.hour < 24:
            val = 0

        distance_day["distance_day"][0]["activities-distance-intraday"][
            "dataset"
        ].append({"time": the_time.strftime("%H:%M:%S"), "value": val})
        the_time = the_time + timedelta(seconds=60)

    return distance_day


def create_syn_data(seed, start_date, end_date):
    """Returns a dict of heart_rate data, activity data, "sleep", "steps","minutesVeryActive", "minutesLightlyActive", "minutesFairlyActive", "distance", "minutesSedentary", "heart_rate_day", "hrv", "distance_day"

    :param start_date: the start date (inclusive) as a string in the format "YYYY-MM-DD"
    :type start_date: str
    :param end_date: the end date (inclusive) as a string in the format "YYYY-MM-DD"
    :type end_date: str
    :return: a dict of heart_rate data, activity data, "sleep", "steps","minutesVeryActive", "minutesLightlyActive", "minutesFairlyActive", "distance", "minutesSedentary", "heart_rate_day", "hrv", "distance_day"
    :rtype: dict
    """

    random.seed(seed)

    num_days = (
        datetime.strptime(end_date, "%Y-%m-%d")
        - datetime.strptime(start_date, "%Y-%m-%d")
    ).days

    # Generate dates as strings
    synth_dates = [
        (datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=i)).strftime(
            "%Y-%m-%d"
        )
        for i in range(num_days)
    ]

    full_dict = {}
    full_dict["sleep"] = []
    full_dict["steps"] = []
    full_dict["minutesVeryActive"] = []
    full_dict["minutesFairlyActive"] = []
    full_dict["minutesLightlyActive"] = []
    full_dict["minutesSedentary"] = []
    full_dict["distance"] = []
    full_dict["heart_rate_day"] = []
    full_dict["hrv"] = []
    full_dict["distance_day"] = []
    full_dict["intraday_breath_rate"] = []
    full_dict["intraday_active_zone_minute"] = []
    full_dict["intraday_heart_rate"] = []
    full_dict["intraday_activity"] = []
    full_dict["intraday_heart_rate"] = []
    full_dict["intraday_hrv"] = []
    full_dict["intraday_spo2"] = []

    for date in synth_dates:
        (
            random_hour,
            random_min,
            random_sec,
            random_duration,
        ) = get_random_sleep_start_time()

        activity = get_activity(date)

        hr = get_heart_rate(date, intraday=False)
        intraday_hr = get_heart_rate(date, intraday=True)

        # Collect all data points

        full_dict["sleep"].append(get_sleep(date))
        full_dict["steps"].append(activity[0])
        full_dict["minutesVeryActive"].append(activity[1])
        full_dict["minutesFairlyActive"].append(activity[2])
        full_dict["minutesLightlyActive"].append(activity[3])
        full_dict["distance"].append(activity[4])
        full_dict["minutesSedentary"].append(activity[5])
        full_dict["heart_rate_day"].append(get_heart_rate(date))
        full_dict["hrv"].append(get_hrv(date))

        full_dict["distance_day"].append(get_distance_day(date))
        full_dict["intraday_breath_rate"].append(get_intraday_breath_rate(date))
        full_dict["intraday_active_zone_minute"].append(
            get_intraday_azm(date, intraday_hr)
        )
        full_dict["intraday_activity"].append(activity[0])
        full_dict["intraday_heart_rate"].append(intraday_hr)
        full_dict["intraday_hrv"].append(
            get_intraday_hrv(date, random_hour, random_min, random_sec, random_duration)
        )
        full_dict["intraday_spo2"].append(
            get_intraday_spo2(
                date, random_hour, random_min, random_sec, random_duration
            )
        )

    # encapsulate to match original data shape
    data = []
    for ele in full_dict["sleep"]:
        data.append(ele)
    full_dict["sleep"] = [{"sleep": data}]

    data = []
    for ele in full_dict["hrv"]:
        data.append(ele)
    full_dict["hrv"] = [{"hrv": data}]

    keys_to_update = [
        "steps",
        "minutesVeryActive",
        "minutesFairlyActive",
        "minutesLightlyActive",
        "distance",
        "minutesSedentary",
    ]

    for key in keys_to_update:
        data = []
        for ele in full_dict[key]:
            data.append(ele)
        full_dict[key] = [{f"activities-{key}": data}]

    full_dict["distance_day"] = full_dict["distance_day"][0]["distance_day"]
    full_dict["heart_rate_day"] = full_dict["heart_rate_day"][0]["heart_rate_day"]

    return full_dict



import os
from datetime import datetime, timedelta
from threading import Lock, Thread

from tqdm import tqdm

__all__ = ["fetch_real_data"]


def fetch_garmin_url(data_type):
    """Fetches the Garmin Connect API endpoint URL corresponding to a given data type.

    This function takes a data type as input and retrieves the associated API endpoint URL from a predefined dictionary.
    If the provided data type is found in the dictionary keys, it returns the respective endpoint URL.
    If the data type is not found, it returns None.

    :param data_type: A string representing the type of data for which the API endpoint URL is requested.
    :type data_type: str
    :return: The Garmin Connect API endpoint URL for the given data type, or None if not found.
    :rtype: Union[str, None]
    """

    url_dict = {
        "steps": "/wellness-service/wellness/dailySummaryChart",
        "body_battery": "/wellness-service/wellness/bodyBattery/reports/daily",
        "hr": "/wellness-service/wellness/dailyHeartRate",
        "blood_pressure": "/bloodpressure-service/bloodpressure/range",
        "floors": "/wellness-service/wellness/floorsChartData/daily",
        "rhr": "/userstats-service/wellness/daily",
        "hydration": "/usersummary-service/usersummary/hydration/daily",
        "sleep": "/wellness-service/wellness/dailySleepData",
        "stress": "/wellness-service/wellness/dailyStress",
        "respiration": "/wellness-service/wellness/daily/respiration",
        "spo2": "/wellness-service/wellness/daily/spo2",
        "hrv": "/hrv-service/hrv",
    }
    if data_type in url_dict.keys():
        return url_dict[data_type]
    else:
        return None


# Steps, HR
def fetch_steps_and_hr(api, data_type, start_date, num_days, params=None):
    display_name = api.display_name
    url = f"{fetch_garmin_url(data_type)}/{display_name}"
    response = []
    for i in tqdm(range(num_days)):
        new_date = datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=i)
        params = {"date": str(new_date.date())}
        response.append(api.connectapi(url, params=params))
    return response


# Floors, Stress, Respiration, Spo2, Hydration, HRV, Training Status, Training Readiness, Activities for Date Aggregated, Day Stress Aggregated
def fetch_aggregated_data(api, data_type, start_date, num_days, params=None):
    response = []
    for i in tqdm(range(num_days)):
        new_date = datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=i)
        url = f"{fetch_garmin_url(data_type)}/{new_date.date()}"
        response.append(api.connectapi(url))
    return response


# Blood Pressure, Weigh Ins
def fetch_blood_pressure_and_weigh_ins(
    api, data_type, start_date, end_date, params=None
):
    url = f"{fetch_garmin_url(data_type)}/{start_date}/{end_date}"
    params = {"includeAll": True}
    return api.connectapi(url, params=params)


# RHR
def fetch_resting_heart_rate(api, data_type, start_date, end_date, params=None):
    display_name = api.display_name
    url = f"{fetch_garmin_url(data_type)}/{display_name}"
    params = {"fromDate": str(start_date), "untilDate": str(end_date), "metricId": 60}
    return api.connectapi(url, params=params)


# Sleep
def fetch_sleep(api, data_type, start_date, num_days, params=None):
    display_name = api.display_name
    response = []
    url = f"{fetch_garmin_url(data_type)}/{display_name}"
    for i in tqdm(range(num_days)):
        new_date = datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=i)
        params = {"date": str(new_date.date()), "nonSleepBufferMinutes": 60}
        response.append(api.connectapi(url, params=params))
    return response


# Body Composition Aggregated, Body Battery
def fetch_body_comp_agg_and_battery(api, data_type, start_date, num_days, params=None):
    response = []
    for i in tqdm(range(num_days)):
        new_date = datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=i)
        url = f"{fetch_garmin_url(data_type)}"
        params = {"startDate": str(new_date), "endDate": str(new_date)}
        response.append(api.connectapi(url, params=params))
    return response


def fetch_real_data(start_date, end_date, data_type, api):
    """Main function for fetching real data from the Garmin Connect API.
    We parallelize this since making requests to the API is day-by-day,
    and API requests are I/O bound.

    :param start_date: the start date represented as a string in the format "YYYY-MM-DD"
    :type start_date: str
    :param end_date: the end date represented as a string in the format "YYYY-MM-DD"
    :type end_date: str
    :param data_type: the type of data to fetch, one of "dates", "steps", "hrs", "brpms"
    :type data_type: str
    :param api: the Garmin Connect API object
    :type api: Garmin
    :return: the data fetched from the API according to the inputs
    :rtype: List
    """

    num_days = (
        datetime.strptime(end_date, "%Y-%m-%d")
        - datetime.strptime(start_date, "%Y-%m-%d")
    ).days
    display_name = api.display_name

    if data_type in ["steps", "hr"]:
        return fetch_steps_and_hr(api, data_type, start_date, num_days)

    elif data_type in ["blood_pressure", "weigh_ins"]:
        return fetch_blood_pressure_and_weigh_ins(api, data_type, start_date, end_date)

    elif data_type in ["rhr"]:
        return fetch_resting_heart_rate(api, data_type, start_date, end_date)

    elif data_type in ["sleep"]:
        return fetch_sleep(api, data_type, start_date, num_days)

    elif data_type in ["body_battery"]:
        return fetch_body_comp_agg_and_battery(api, data_type, start_date, num_days)

    elif data_type in ["floors", "stress", "respiration", "spo2", "hydration", "hrv"]:
        return fetch_aggregated_data(api, data_type, start_date, num_days)

    elif data_type == "dates":
        return [
            datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=i)
            for i in tqdm(range(num_days))
        ]

    return None



from datetime import datetime, time, timedelta

import numpy as np

from ...utils import bin_search, seed_everything
from ..device import BaseDevice
from .fitbit_authenticate import fitbit_application
from .fitbit_sense_fetch import fetch_real_data
from .fitbit_sense_gen import create_syn_data


class FitbitCharge6(BaseDevice):
    """This device allows you to work with data from the `Fitbit charge  <(https://www.fitbit.com/global/au/products/trackers/charge6)>`_ device.
    Available datatypes for this device are:

    * `intraday_breath_rate`: collected per stage of sleep
    * `intraday_active_zone_minute`: collected per minute
    * `intraday_activity`: collected per minute
    * `intraday_heart_rate`: collected per second
    * `intraday_hrv`: rmssd, lf and hf and collected during sleep
    * `intraday_spo2`: in percentage collected during sleep
    * `sleep`: sleep data
    * `steps`: steps data
    * `minutesVeryActive`: number of minutes with high activity
    * `minutesLightlyActive`: number of minutes with light activity
    * `minutesFairlyActive`: number of minutes with fair activity
    * `distance`: in miles
    * `minutesSedentary`: number of minutes with no activity

    :param seed: random seed for synthetic data generation, defaults to 0
    :type seed: int, optional
    :param synthetic_start_date: start date for synthetic data generation, defaults to "2024-04-01"
    :type synthetic_start_date: str, optional
    :param synthetic_end_date: end date for synthetic data generation, defaults to "2024-05-13"
    :type synthetic_end_date: str, optional
    """

    name = "fitbit/fitbit_charge_6"

    def __init__(
        self,
        seed=0,
        synthetic_start_date="2024-01-01",
        synthetic_end_date="2024-01-31",
    ):

        params = {
            "seed": seed,
            "synthetic_start_date": synthetic_start_date,
            "synthetic_end_date": synthetic_end_date,
        }

        self._initialize_device_params(
            [
                "intraday_breath_rate",
                "intraday_active_zone_minute",
                "intraday_activity",
                "intraday_heart_rate",
                "intraday_hrv",
                "intraday_spo2",
                "sleep",
                "steps",
                "minutesVeryActive",
                "minutesLightlyActive",
                "minutesFairlyActive",
                "distance",
                "minutesSedentary",
            ],
            params,
            {
                "seed": 0,
                "synthetic_start_date": "2024-01-01",
                "synthetic_end_date": "2024-01-31",
            },
        )

    def _default_params(self):
        return {
            "start_date": "2022-04-24",
            "end_date": "2022-04-28",
        }

    def _filter_synthetic(self, data, data_type, params):

        date_format = "%Y-%m-%d"
        date1 = datetime.strptime(self.init_params["synthetic_start_date"], date_format)
        date2 = datetime.strptime(params["start_date"], date_format)

        date3 = datetime.strptime(self.init_params["synthetic_end_date"], date_format)
        date4 = datetime.strptime(params["end_date"], date_format)

        delta1 = date2 - date1
        delta2 = date3 - date4

        num_days_start = delta1.days
        num_days_end = delta2.days

        key_map = {
            "sleep": "sleep",
            "steps": "activities-steps",
            "minutesVeryActive": "activities-minutesVeryActive",
            "minutesLightlyActive": "activities-minutesLightlyActive",
            "minutesFairlyActive": "activities-minutesFairlyActive",
            "distance": "activities-distance",
            "minutesSedentary": "activities-minutesSedentary",
            "hrv": "hrv",
        }

        if data_type in key_map:
            key = key_map[data_type]
            if key is None:
                return data
            intermediary = data[0][key][num_days_start : -num_days_end + 1]
            return [{key: intermediary}]
        return data

    def _get_real(self, data_type, params):
        data = fetch_real_data(
            data_type,
            self.user,
            start_date=params["start_date"],
            end_date=params["end_date"],
        )
        return data

    def _gen_synthetic(self):

        syn_data = create_syn_data(
            0,
            self.init_params["synthetic_start_date"],
            self.init_params["synthetic_end_date"],
        )

        self.intraday_breath_rate = syn_data["intraday_breath_rate"]
        self.intraday_activity = syn_data["intraday_activity"]
        self.intraday_active_zone_minute = syn_data["intraday_active_zone_minute"]
        self.intraday_heart_rate = syn_data["intraday_heart_rate"]
        self.intraday_hrv = syn_data["intraday_hrv"]
        self.intraday_spo2 = syn_data["intraday_spo2"]
        self.sleep = syn_data["sleep"]
        self.steps = syn_data["steps"]
        self.minutesVeryActive = syn_data["minutesVeryActive"]
        self.minutesFairlyActive = syn_data["minutesFairlyActive"]
        self.minutesLightlyActive = syn_data["minutesLightlyActive"]
        self.distance = syn_data["distance"]
        self.minutesSedentary = syn_data["minutesSedentary"]

    def _authenticate(self, token=""):
        if token == "":
            self.user = fitbit_application()
        else:
            self.user = token



import http
import json

import numpy as np
import pandas as pd

from ...devices.device import BaseDevice
from ...utils import bin_search, seed_everything
from .pro_cgm_fetch import dexcom_authenticate, fetch_data, refresh_access_token
from .pro_cgm_gen import create_synth


class DexcomProCGM(BaseDevice):
    """This device allows you to work with data from the `Dexcom Pro CGM <https://provider.dexcom.com/products/dexcom-g6-pro>`_ device.
    Available datatypes for this device are:

    * `data`: contains all data in one dictionary

    :param seed: random seed for synthetic data generation, defaults to 0
    :type seed: int, optional
    :param synthetic_start_date: start date for synthetic data generation, defaults to "2022-03-01"
    :type synthetic_start_date: str, optional
    :param synthetic_end_date: end date for synthetic data generation, defaults to "2022-06-17"
    :type synthetic_end_date: str, optional
    """

    name = "dexcom/pro_cgm"

    def __init__(
        self, seed=0, synthetic_start_date="2022-02-16", synthetic_end_date="2022-05-15"
    ):

        params = {
            "seed": seed,
            "synthetic_start_date": synthetic_start_date,
            "synthetic_end_date": synthetic_end_date,
        }

        self._initialize_device_params(
            ["data"],
            params,
            {
                "seed": 0,
                "synthetic_start_date": "2022-02-16",
                "synthetic_end_date": "2022-05-15",
            },
        )

    def _default_params(self):
        return {
            "start_date": self.init_params["synthetic_start_date"],
            "end_date": self.init_params["synthetic_end_date"],
        }

    def _get_real(self, data_type, params):
        # there is really only one data type for this device,
        # so we don't need to check the data_type

        return fetch_data(
            self.access_token,
            start_date=params["start_date"],
            end_date=params["end_date"],
        )

    def _filter_synthetic(self, data, data_type, params):
        # there is really only one data type for this device,
        # so we don't need to check the data_type

        start_ts = pd.Timestamp(params["start_date"])
        end_ts = pd.Timestamp(params["end_date"])

        timestamps = [pd.Timestamp(x["systemTime"]) for x in self.data["egvs"]]

        start_idx = bin_search(timestamps, start_ts)
        end_idx = bin_search(timestamps, end_ts)

        return {
            "unit": "mg/dL",
            "rateUnit": "mg/dL/min",
            "egvs": self.data["egvs"][start_idx:end_idx],
        }

    def _gen_synthetic(self):
        # generate random data according to seed
        seed_everything(self.init_params["seed"])

        self.data = create_synth(
            self.init_params["synthetic_start_date"],
            self.init_params["synthetic_end_date"],
        )

    def _authenticate(self, auth_creds, use_cache=True):
        if use_cache and hasattr(self, "access_token"):
            return

        if "refresh_token" in auth_creds:
            self.refresh_token, self.access_token = refresh_access_token(
                auth_creds["refresh_token"],
                auth_creds["client_id"],
                auth_creds["client_secret"],
            )
        else:
            self.refresh_token, self.access_token = dexcom_authenticate(
                auth_creds["client_id"], auth_creds["client_secret"]
            )



import random
from datetime import datetime, timedelta

import numpy as np


def get_hrv_data(start_date, num_days):
    """
    Generate synthetic Heart Rate Variability (HRV) data for a specified date range.

    This function creates synthetic HRV data for a given number of days starting from a specified date.
    It includes various HRV metrics such as last night's average, last night's 5-minute high, baseline values,
    status, feedback phrases, and associated timestamps.

    :param start_date: The start date for the HRV data generation in "YYYY-MM-DD" format.
    :type start_date: str
    :param num_days: The number of days to generate HRV data for, starting from the start_date.
    :type num_days: int
    :return: A list of dictionaries, each containing HRV data for a specific day. This includes HRV summary,
        readings, timestamps, and sleep-related timestamps.
    :rtype: List[Dict]
    """

    hrv_data = []
    start_date_obj = datetime.strptime(start_date, "%Y-%m-%d")

    for i in range(num_days):
        date = start_date_obj + timedelta(days=i)
        date_str = date.strftime("%Y-%m-%d")
        start_timestamp_gmt = f"{date_str}T06:00:00.0"
        end_timestamp_gmt = (
            f"{date_str}T13:{random.randint(0, 59):02d}:{random.randint(0, 59):02d}.0"
        )
        start_timestamp_local = f"{(datetime.strptime(start_date, '%Y-%m-%d') - timedelta(days=1)).strftime('%Y-%m-%d')}T23:00:00.0"
        end_timestamp_local = (
            f"{date_str}T06:{random.randint(0, 59):02d}:{random.randint(0, 59):02d}.0"
        )

        last_night_avg = random.randint(15, 30)
        last_night_5min_high = random.randint(30, 60)
        baseline = {
            "lowUpper": random.randint(15, 20),
            "balancedLow": random.randint(20, 25),
            "balancedUpper": random.randint(25, 35),
            "markerValue": round(random.uniform(0.3, 0.6), 8),
        }
        status = random.choice(["BALANCED", "ELEVATED", "LOW"])
        feedback_phrase = f"HRV_{status}_RANDOM"

        hrv_entry = {
            "userProfilePk": random.randint(10000000, 99999999),
            "hrvSummary": {
                "calendarDate": date_str,
                "weeklyAvg": None,
                "lastNightAvg": last_night_avg,
                "lastNight5MinHigh": last_night_5min_high,
                "baseline": baseline,
                "status": status,
                "feedbackPhrase": feedback_phrase,
                "createTimeStamp": date.strftime("%Y-%m-%dT%H:%M:%S.000"),
            },
            "hrvReadings": [],
            "startTimestampGMT": start_timestamp_gmt,
            "endTimestampGMT": end_timestamp_gmt,
            "startTimestampLocal": start_timestamp_local,
            "endTimestampLocal": end_timestamp_local,
            "sleepStartTimestampGMT": None,
            "sleepEndTimestampGMT": None,
            "sleepStartTimestampLocal": None,
            "sleepEndTimestampLocal": None,
        }

        hrv_data.append(hrv_entry)

    return hrv_data


def get_steps_data(start_date, num_days):
    """
    Generate synthetic step data for a specified date range.

    This function creates synthetic step data for each 15-minute interval within a specified date range.
    It includes the number of steps taken, activity level, and a boolean indicating if the activity level is constant.

    :param start_date: The start date for generating step data in "YYYY-MM-DD" format.
    :type start_date: str
    :param num_days: The number of days to generate step data for, starting from the start_date.
    :type num_days: int
    :return: A list of lists, where each inner list contains dictionaries of step data for each 15-minute
        interval in a day. Each dictionary includes the start and end times for the interval, steps taken,
        activity level, and a constant activity level indicator.
    :rtype: List[List[Dict]]
    """
    steps_data = []

    for _ in range(num_days):
        calendar_date = (
            datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=_)
        ).strftime("%Y-%m-%d")
        start_timestamp_gmt = f"{calendar_date}T06:00:00.0"

        steps_day = []
        for i in range(96):  # 24 hours * 60 minutes / 15 minutes
            interval_start = datetime.strptime(
                start_timestamp_gmt, "%Y-%m-%dT%H:%M:%S.0"
            ) + timedelta(minutes=15 * i)
            interval_end = interval_start + timedelta(minutes=15)
            steps = int(np.random.normal(90, 30))
            steps_entry = {
                "startGMT": interval_start.strftime("%Y-%m-%dT%H:%M:%S.0"),
                "endGMT": interval_end.strftime("%Y-%m-%dT%H:%M:%S.0"),
                "steps": max(steps, 0),
                "pushes": 0,
                "primaryActivityLevel": random.choice(
                    ["active", "sedentary", "sleeping", "none"]
                ),
                "activityLevelConstant": random.choice([True, False]),
            }
            steps_day.append(steps_entry)
        steps_data.append(steps_day)

    return steps_data


def get_heart_rate_data(start_date, num_days, steps_data):
    """Generate synthetic heart rate data for a specified date range.

    This function generates synthetic heart rate data for a given date range, including various heart rate metrics
    such as resting heart rate, maximum heart rate, minimum heart rate, and additional descriptors and values.

    :param start_date: The start date in the format "YYYY-MM-DD".
    :type start_date: str
    :param num_days: The number of days for which to generate heart rate data.
    :type num_days: int
    :return: A list of dictionaries, each containing heart rate data for a specific day, including resting heart rate,
        maximum and minimum heart rate, and additional heart rate descriptors and values.
    :rtype: List[Dict]
    """
    heart_rate_data = []

    for _ in range(num_days):
        calendar_date = (
            datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=_)
        ).strftime("%Y-%m-%d")
        start_timestamp_gmt = f"{calendar_date}T07:00:00.0"
        end_timestamp_gmt = (
            datetime.strptime(start_timestamp_gmt, "%Y-%m-%dT%H:%M:%S.0")
            + timedelta(days=1)
        ).strftime("%Y-%m-%dT%H:%M:%S.0")
        start_timestamp_local = f"{calendar_date}T00:00:00.0"
        end_timestamp_local = (
            datetime.strptime(start_timestamp_local, "%Y-%m-%dT%H:%M:%S.0")
            + timedelta(days=1)
        ).strftime("%Y-%m-%dT%H:%M:%S.0")
        resting_heart_rate = random.randint(50, 90)
        max_heart_rate = random.randint(resting_heart_rate + 5, 120)
        min_heart_rate = random.randint(resting_heart_rate + 2, max_heart_rate - 1)

        last_seven_days_avg = random.randint(
            resting_heart_rate - 2, resting_heart_rate + 2
        )

        heart_rate_values = []
        start_datetime = datetime.strptime(start_timestamp_gmt, "%Y-%m-%dT%H:%M:%S.0")
        end_datetime = datetime.strptime(end_timestamp_gmt, "%Y-%m-%dT%H:%M:%S.0")
        while start_datetime < end_datetime:
            steps_arrdict_day = steps_data[_]
            step_timestamps = np.array(
                [
                    datetime.strptime(x["startGMT"], "%Y-%m-%dT%H:%M:%S.0").timestamp()
                    for x in steps_arrdict_day
                ]
            )
            step_idx = np.where(
                np.logical_and(
                    datetime.strptime(calendar_date, "%Y-%m-%d").timestamp()
                    + 6 * 60 * 60
                    >= step_timestamps,
                    datetime.strptime(calendar_date, "%Y-%m-%d").timestamp()
                    <= step_timestamps + 15 * 60,
                )
            )[0][0]
            step_val_avg = steps_arrdict_day[step_idx]["steps"]
            heart_rate = int(step_val_avg * 0.5 + 75 + np.random.randn() * 10)
            timestamp = int(
                start_datetime.timestamp() * 1000
            )  # Convert to milliseconds
            heart_rate_values.append([timestamp, heart_rate])
            start_datetime += timedelta(minutes=60)

        heart_rate_entry = {
            "userProfilePK": random.randint(10000000, 99999999),
            "calendarDate": calendar_date,
            "startTimestampGMT": start_timestamp_gmt,
            "endTimestampGMT": end_timestamp_gmt,
            "startTimestampLocal": start_timestamp_local,
            "endTimestampLocal": end_timestamp_local,
            "maxHeartRate": max_heart_rate,
            "minHeartRate": min_heart_rate,
            "restingHeartRate": resting_heart_rate,
            "lastSevenDaysAvgRestingHeartRate": last_seven_days_avg,
            "heartRateValueDescriptors": [
                {"key": "timestamp", "index": 0},
                {"key": "heartrate", "index": 1},
            ],
            "heartRateValues": heart_rate_values,
        }

        heart_rate_data.append(heart_rate_entry)

    return heart_rate_data


def get_body_battery_data(start_date, num_days):
    """
    Generate synthetic body battery data for a specified number of days.

    This function creates synthetic data representing the body battery level for each day within the given date range.
    The data includes the starting and ending timestamps, charged and drained values, and an array of body battery levels
    at different times throughout the day. The body battery levels and their corresponding timestamps are randomly generated.

    :param start_date: The start date for generating body battery data in "YYYY-MM-DD" format.
    :type start_date: str
    :param num_days: The number of days for which to generate body battery data, starting from the start_date.
    :type num_days: int
    :return: A list where each item is a list of dictionary containing the date, charged and drained values, timestamps,
        and an array of body battery levels for each time interval within the day.
    :rtype: List[List[Dict]]
    """

    body_battery_data = []

    for day in range(num_days):
        calendar_date = (
            datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=day)
        ).strftime("%Y-%m-%d")
        start_timestamp_gmt = f"{calendar_date}T23:00:00.0"
        end_timestamp_gmt = (
            datetime.strptime(start_timestamp_gmt, "%Y-%m-%dT%H:%M:%S.0")
            + timedelta(days=1)
        ).strftime("%Y-%m-%dT%H:%M:%S.0")
        start_timestamp_local = f"{calendar_date}T00:00:00.0"
        end_timestamp_local = (
            datetime.strptime(start_timestamp_local, "%Y-%m-%dT%H:%M:%S.0")
            + timedelta(days=1)
        ).strftime("%Y-%m-%dT%H:%M:%S.0")

        charged = random.randint(0, 100)
        drained = random.randint(0, 100)

        body_battery_values_array = []
        start_datetime = datetime.strptime(start_timestamp_gmt, "%Y-%m-%dT%H:%M:%S.0")
        end_datetime = datetime.strptime(end_timestamp_gmt, "%Y-%m-%dT%H:%M:%S.0")
        while start_datetime < end_datetime:
            timestamp = int(
                start_datetime.timestamp() * 1000
            )  # Convert to milliseconds
            body_battery_level = random.randint(0, 100)  # Random body battery level
            body_battery_values_array.append([timestamp, body_battery_level])
            # Assuming 15-minute intervals
            start_datetime += timedelta(minutes=15)

        body_battery_entry = [
            {
                "date": calendar_date,
                "charged": charged,
                "drained": drained,
                "startTimestampGMT": start_timestamp_gmt,
                "endTimestampGMT": end_timestamp_gmt,
                "startTimestampLocal": start_timestamp_local,
                "endTimestampLocal": end_timestamp_local,
                "bodyBatteryValuesArray": body_battery_values_array,
                "bodyBatteryValueDescriptorDTOList": [
                    {
                        "bodyBatteryValueDescriptorIndex": 0,
                        "bodyBatteryValueDescriptorKey": "timestamp",
                    },
                    {
                        "bodyBatteryValueDescriptorIndex": 1,
                        "bodyBatteryValueDescriptorKey": "bodyBatteryLevel",
                    },
                ],
            }
        ]

        body_battery_data.append(body_battery_entry)

    return body_battery_data



import os
import pickle
import webbrowser
from datetime import datetime
from urllib.parse import urlencode

import requests

from ...utils import seed_everything
from ..device import BaseDevice
from .evo_fetch import fetch_real_data
from .evo_gen import create_syn_data


class EVO(BaseDevice):

    name = "biostrap/evo"

    def __init__(self, seed=0, start_date="2023-06-05", end_date="2023-06-20"):
        params = {
            "seed": seed,
            "start_date": str(start_date),
            "end_date": str(end_date),
        }

        self._initialize_device_params(
            [
                "activities",
                "bpm",
                "brpm",
                "hrv",
                "spo2",
                "rest_cals",
                "work_cals",
                "active_cals",
                "step_cals",
                "total_cals",
                "sleep_session",
                "sleep_detail",
                "steps",
                "distance",
            ],
            params,
            {
                "seed": 0,
                "synthetic_start_date": "2023-06-05",
                "synthetic_end_date": "2023-06-20",
            },
        )

    def _default_params(self):
        return {
            "start_date": self.init_params["synthetic_start_date"],
            "end_date": self.init_params["synthetic_end_date"],
        }

    def _get_real(self, data_type, params):
        return fetch_real_data(
            self.access_token, params["start_date"], params["end_date"], data_type
        )

    def _filter_synthetic(self, data, data_type, params):
        start_date = params["start_date"]
        end_date = params["end_date"]

        start_datetime = f"{start_date} 00:00:00"
        end_datetime = f"{end_date} 23:59:59"

        # For data types that are stored with a date string as a key
        if data_type in [
            "rest_cals",
            "work_cals",
            "active_cals",
            "step_cals",
            "total_cals",
        ]:
            return {
                date: value
                for date, value in data.items()
                if start_date <= date <= end_date
            }

        # For data types that are stored with a datetime string as a key in a tuple
        elif data_type in [
            "bpm",
            "brpm",
            "spo2",
            "hrv",
        ]:
            return {
                key: value
                for key, value in data.items()
                if start_datetime <= key[0] <= end_datetime
            }

        # For steps and distance that use datetime strings as keys (assuming distance also has datetime as per steps)
        elif data_type in ["steps", "distance"]:
            return {
                datetime: value
                for datetime, value in data.items()
                if start_datetime <= datetime <= end_datetime
            }

        else:
            return data

    def _gen_synthetic(self):
        # generate random data according to seed
        seed_everything(self.init_params["seed"])
        # and based on start and end dates
        (
            self.activities,
            self.bpm,
            self.brpm,
            self.hrv,
            self.spo2,
            self.rest_cals,
            self.work_cals,
            self.active_cals,
            self.step_cals,
            self.total_cals,
            self.sleep_session,
            self.sleep_detail,
            self.steps,
            self.distance,
        ) = create_syn_data(
            self.init_params["synthetic_start_date"],
            self.init_params["synthetic_end_date"],
        )

    # We get the access token to make requests to the Biostrap API
    def _authenticate(self, auth_creds):
        self.client_id = auth_creds["client_id"]
        self.client_secret = auth_creds["client_secret"]
        redirect_uri = "https://127.0.0.1:8080"
        token_url = "https://auth.biostrap.com/token"

        # Generate the authorization URL and open it in the default web browser
        params = {
            "client_id": self.client_id,
            "response_type": "code",
            "redirect_uri": redirect_uri,
        }
        authorization_url = f"https://auth.biostrap.com/authorize?{urlencode(params)}"

        # Instruct the user to open the URL in their browser ()
        print(
            f"Please go to the following URL and authorize access: {authorization_url}"
        )
        print(
            "If a page cannot be displayed, please make sure you are logged into the Biostrap account in your browser."
        )

        # Get the authorization response URL from the command line (the method Jack and I talked about didn't work)
        authorization_response = input("Enter the full callback URL: ")

        code = authorization_response.split("code=")[1].split("&")[0]

        # Now we can request the access token
        data = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": redirect_uri,
        }
        auth = (self.client_id, self.client_secret)
        response = requests.post(token_url, auth=auth, data=data)

        # Get the access token from the response
        token_json = response.json()
        self.access_token = token_json["access_token"]



from datetime import datetime, timedelta

from .fenix_gen_1 import (
    get_body_battery_data,
    get_heart_rate_data,
    get_hrv_data,
    get_steps_data,
)
from .fenix_gen_2 import (
    get_blood_pressure_data,
    get_floors_data,
    get_hydration_data,
    get_resting_hr_data,
)
from .fenix_gen_3 import (
    get_respiration_data,
    get_sleep_data,
    get_spo2_data,
    get_stress_data,
)

__all__ = ["create_syn_data"]


def create_syn_data(start_date, end_date):
    """
    Returns a dictionary of synthetic health and activity data for a specified date range.
    :param start_date: the start date (inclusive) as a string in the format "YYYY-MM-DD"
    :type start_date: str
    :param end_date: the end date (inclusive) as a string in the format "YYYY-MM-DD"
    :type end_date: str
    :return: A dictionary containing synthetic data for various health and activity metrics, each
        element is a list or dictionary representing data for a specific day.
    :rtype: Dict
    """

    num_days = (
        datetime.strptime(end_date, "%Y-%m-%d")
        - datetime.strptime(start_date, "%Y-%m-%d")
    ).days

    synth_data = {
        "dates": [
            datetime.strptime(start_date, "%Y-%m-%d") + timedelta(days=i)
            for i in range(num_days)
        ],
        "hrv": get_hrv_data(start_date, num_days),
        "steps": get_steps_data(start_date, num_days),
        "hr": None,
        "body_battery": get_body_battery_data(start_date, num_days),
        "blood_pressure": get_blood_pressure_data(start_date, end_date, 100),
        "floors": get_floors_data(start_date, num_days),
        "rhr": get_resting_hr_data(start_date, num_days),
        "hydration": get_hydration_data(start_date, num_days),
        "sleep": get_sleep_data(start_date, num_days),
        "stress": get_stress_data(start_date, num_days),
        "respiration": get_respiration_data(start_date, num_days),
        "spo2": get_spo2_data(start_date, num_days),
    }

    synth_data["hr"] = get_heart_rate_data(start_date, num_days, synth_data["steps"])

    return synth_data



import random

import numpy as np

__all__ = ["is_notebook", "seed_everything"]


def is_notebook() -> bool:
    """Check if we are running in a notebook.

    :return: `True` if we are running in a notebook, `False` otherwise.
    :rtype: bool
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def seed_everything(seed):
    """Set random seed for reproducibility.

    :param seed: the seed to use
    :type seed: int
    """
    np.random.seed(seed)
    random.seed(seed)


def bin_search_aux(data, start, end, target):
    """Binary search for a target in a sorted array.
    This is a helper function for `bin_search`.

    :param data: the sorted array
    :type data: list
    :param start: the start index
    :type start: int
    :param end: the end index
    :type end: int
    :param target: the target to search for
    :type target: int
    :return: the index of the target
    :rtype: int
    """
    if start >= end:
        return start

    mid = (start + end) // 2
    if data[mid] == target:
        return mid
    elif data[mid] < target:
        return bin_search_aux(data, mid + 1, end, target)
    else:
        return bin_search_aux(data, start, mid - 1, target)


def bin_search(data, target):
    """Binary search for a target in a sorted array.

    :param data: the sorted array
    :type data: list
    :param target: the target to search for
    :type target: int
    :return: the index of the target
    :rtype: int
    """

    return bin_search_aux(data, 0, len(data) - 1, target)



import os
from datetime import datetime, timedelta
from pathlib import Path

import numpy as np
import pandas as pd
import wget

from ...devices.device import BaseDevice
from ...utils import bin_search, seed_everything
from .withings_authenticate import refresh_access_token, withings_authenticate
from .withings_extract import fetch_measurements
from .withings_gen import create_syn_bodyplus


class BodyPlus(BaseDevice):
    """This device allows you to work with data from the `Withings Body+ <https://www.withings.com/us/en/body-plus>`_ device.
    Available datatypes for this device are:

    * `measurements`: contains all measurements made with the scale

    :param seed: random seed for synthetic data generation, defaults to 0
    :type seed: int, optional
    :param synthetic_start_date: start date for synthetic data generation, defaults to "2021-06-01"
    :type start_date: str, optional
    """

    name = "withings/bodyplus"

    def __init__(self, seed=0, synthetic_start_date="2021-06-01"):

        params = {"seed": seed, "synthetic_start_date": synthetic_start_date}

        self._initialize_device_params(
            ["measurements"],
            params,
            {
                "seed": 0,
                "synthetic_start_date": "2021-06-01",
            },
        )

    def _default_params(self):
        return {
            "start": self.init_params["synthetic_start_date"],
            "end": datetime.strftime(
                datetime.strptime(self.init_params["synthetic_start_date"], "%Y-%m-%d")
                + timedelta(days=900),
                "%Y-%m-%d",
            ),
        }

    def _get_real(self, data_type, params):
        start = datetime.strptime(params["start"], "%Y-%m-%d")
        end = datetime.strptime(params["end"], "%Y-%m-%d")

        return fetch_measurements(self.access_token, start, end)

    def _filter_synthetic(self, data, data_type, params):
        start_ts = pd.Timestamp(params["start"])
        end_ts = pd.Timestamp(params["end"])

        start_idx = bin_search(np.array(data.date), start_ts)
        end_idx = bin_search(np.array(data.date), end_ts)

        return data.iloc[start_idx:end_idx]

    def _gen_synthetic(self):
        # generate random data according to seed
        seed_everything(self.init_params["seed"])

        self.measurements = create_syn_bodyplus(
            self.init_params["synthetic_start_date"]
        )

    def _authenticate(self, auth_creds):
        if "access_token" in auth_creds:
            self.access_token = auth_creds["access_token"]
            return

        if "refresh_token" in auth_creds:
            self.refresh_token, self.access_token = refresh_access_token(
                auth_creds["refresh_token"],
                auth_creds["client_id"],
                auth_creds["client_secret"],
            )
        else:
            self.refresh_token, self.access_token = withings_authenticate(
                auth_creds["client_id"], auth_creds["client_secret"]
            )



import http
import json
import time
import urllib
from datetime import datetime

import pandas as pd
import requests

__all__ = ["refresh_access_token", "dexcom_authenticate", "fetch_data"]


def refresh_access_token(refresh_token, client_id, client_secret):
    # gives us access token given the refresh token

    params = {
        "client_id": client_id,
        "client_secret": client_secret,
        "refresh_token": refresh_token,
        "grant_type": "refresh_token",
        "redirect_uri": "https://www.google.com",
    }

    out = requests.post("https://api.dexcom.com/v2/oauth2/token", data=params)

    body = json.loads(out.text)

    return body["refresh_token"], body["access_token"]


def dexcom_authenticate(your_client_id, your_client_secret):
    your_redirect_uri = "https://www.google.com"
    your_state_value = "1234"

    url = f"https://api.dexcom.com/v2/oauth2/login?client_id={your_client_id}&redirect_uri={your_redirect_uri}&response_type=code&scope=offline_access&state={your_state_value}"

    print(url)

    print("redirect url below:")
    time.sleep(0.1)
    redirect_url = input(">")

    try:
        your_authorization_code = urllib.parse.parse_qs(
            urllib.parse.urlparse(redirect_url).query
        )["code"][0]
    except Exception as e:
        exception_str = f"Caught error:\n{e}\n"
        exception_str += "\nPlease copy and paste the entire URL (including https)"
        raise Exception(exception_str)

    conn = http.client.HTTPSConnection("api.dexcom.com")

    payload = f"client_secret={your_client_secret}&client_id={your_client_id}&code={your_authorization_code}&grant_type=authorization_code&redirect_uri={your_redirect_uri}"

    headers = {
        "content-type": "application/x-www-form-urlencoded",
        "cache-control": "no-cache",
    }

    conn.request("POST", "/v2/oauth2/token", payload, headers)

    res = conn.getresponse()
    data = res.read()

    json_response = json.loads(data.decode("utf-8"))

    if "error" in json_response.keys() and json_response["error"] == "invalid_grant":
        exception_str = (
            "The code you got has expired.\nAuthorize and enter the redirect URL again."
        )
        raise Exception(exception_str)
    else:
        body = json.loads(data.decode("utf-8"))
        access_token = body["access_token"]
        refresh_token = body["refresh_token"]

        print(f'Entire response was {data.decode("utf-8")}')
        print(f"Our access token is {access_token}")

    return refresh_token, access_token


def fetch_data(access_token, start_date="2022-02-16", end_date="2022-05-15"):
    start_date = start_date + "T15:30:00"
    end_date = end_date + "T15:45:00"

    headers = {"authorization": f"Bearer {access_token}"}

    endpoint = f"https://api.dexcom.com/v2/users/self/egvs?startDate={start_date}&endDate={end_date}"

    out = json.loads(requests.get(endpoint, headers=headers).text)

    if "errors" in out.keys():
        exception_str = (
            f'Got error(s) {out["errors"]}. Fix start and end dates and rerun.'
        )
        raise Exception(exception_str)
    elif "fault" in out.keys():
        exception_str = (
            f'Got fault {out["fault"]}. You might need to request another access token.'
        )
        raise Exception(exception_str)
    else:

        return out
